# Macaroon - a desktop macro tool
# Copyright (C) 2007 Eitan Isaacson <eitan@ascender.com>
# All rights reserved.

# This file may be distributed and/or modified under the terms of
# the GNU General Public License version 2 as published by
# the Free Software Foundation.
# This file is distributed without any warranty; without even the implied
# warranty of merchantability or fitness for a particular purpose.
# See "COPYING" in the source distribution for more information.

# Headers in this file shall remain intact.

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Pango
from gi.repository import GtkSource

from . import script_factory
from queue import Queue
from macaroon.playback import MacroSequence

APP_ID = None

MacroSequence.startReally = MacroSequence.start
MacroSequence.start = lambda x: None

from .about import MacaroonAboutDialog

import pyatspi

_ = lambda x: x


class Main:
  start_tooltip = _('Record new keyboard macro')
  stop_tooltip = _('Stop recording macro')
  def __init__(self):
    status_icon = Gtk.StatusIcon.new_from_stock(Gtk.STOCK_MEDIA_RECORD)
    status_icon.connect('activate', self._onActivate)
    status_icon.connect('popup-menu', self._onPopup)
    status_icon.set_tooltip_text(self.start_tooltip)
    self.ui_manager = self._newUIManager()
    self.script_buffer = ScriptBuffer(self.ui_manager)
    self.script_buffer.clearBuffer()
    self.script_buffer.connect('notify::recording',
                               self._onRecordChange, status_icon)
    self.macro_preview = None
    # Get program ID
    global APP_ID
#    from gnome import program_get
#    _prog = program_get()
#    if _prog is not None:
#      APP_ID = _prog.get_app_id()

    pyatspi.Registry.start()

  def _onRecordChange(self, gobject, pspec, status_icon):
    is_recording = self.script_buffer.get_property('recording')
    if is_recording:
      status_icon.set_from_stock(Gtk.STOCK_MEDIA_STOP)
      status_icon.set_tooltip_text(self.stop_tooltip)
    else:
      status_icon.set_from_stock(Gtk.STOCK_MEDIA_RECORD)
      status_icon.set_tooltip_text(self.start_tooltip)

  def _onActivate(self, status_icon):
    is_recording = self.script_buffer.get_property('recording')
    if not is_recording:
      self.script_buffer.startRecord()
    else:
      self.script_buffer.stopRecord()
      if self.macro_preview is None:
        self.macro_preview = MacroPreview(self.script_buffer)
      self.macro_preview.show_all()

  def _newUIManager(self):
    popup_ui = '''
<ui>
<popup>
    <placeholder name="ScriptType">
    </placeholder>
    <separator />
    <menuitem action="About" />
    <separator />
    <menuitem action="Quit" />
</popup>
</ui>'''
    main_action_group = Gtk.ActionGroup('MenuActions')
    main_action_group.add_actions([
        ('ScriptType', None, 'Script type'),
        ('Quit', Gtk.STOCK_QUIT, _('_Quit'), None, None, self._onQuit),
        ('About', Gtk.STOCK_ABOUT, _('_About'), None, None, self._onAbout)])
    ui_manager = Gtk.UIManager()
    ui_manager.add_ui_from_string(popup_ui)
    ui_manager.insert_action_group(main_action_group, 0)
    return ui_manager

  def _onPopup(self, status_icon, button, activate_time):
    menu = self.ui_manager.get_widget('/popup')
    menu.popup(None, None, Gtk.StatusIcon.position_menu,
               status_icon, button, activate_time)

  def _onQuit(self, action):
    pyatspi.Registry.stop()

  def _onAbout(self, action):
    '''
    Shows the about dialog.

    @param widget: The widget that emitted the signal that callback caught.
    @type widget: L{Gtk.Widget}
    '''
    about = MacaroonAboutDialog()
    about.show_all()


class MacroPreview(Gtk.Window):
  def __init__(self, script_buffer):
    GObject.GObject.__init__(self)
    self.set_title(_('Macro preview'))
    self.set_default_size(480, 720)
    self.set_border_width(6)
    self.connect('delete-event', self._onDelete)
    self.script_buffer = script_buffer
    text_view = GtkSource.View.new_with_buffer(self.script_buffer)
    text_view.set_editable(True)
    text_view.set_cursor_visible(True)
    text_view.modify_font(Pango.FontDescription('Mono'))
    sw = Gtk.ScrolledWindow()
    sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
    sw.set_shadow_type(Gtk.ShadowType.IN)
    sw.add(text_view)
    vbox = Gtk.VBox()
    vbox.set_spacing(3)
    vbox.pack_start(sw, True, True, 0)
    bbox = Gtk.HButtonBox()
    bbox.set_layout(Gtk.ButtonBoxStyle.START)
    for label, callback in [(Gtk.STOCK_SAVE_AS, self._onSave),
                            (Gtk.STOCK_CLEAR, self._onClear),
                            (Gtk.STOCK_MEDIA_RECORD, self._onRecord),
                            (Gtk.STOCK_MEDIA_PLAY, self._onPlay)]:
      button = Gtk.Button(label)
      button.set_use_stock(True)
      button.connect('clicked', callback)
      bbox.pack_start(button, True, True, 0)
      if label == Gtk.STOCK_MEDIA_RECORD:
        self.script_buffer.connect('notify::recording',
                                   self._onRecordChange, button)
    vbox.pack_start(bbox, False, False, 0)
    self.progress_bar = Gtk.ProgressBar()
    vbox.pack_start(self.progress_bar, False, False, 0)
    self.add(vbox)

  def _onPlay(self, button):
    script = self.script_buffer.get_text(self.script_buffer.get_start_iter(),
                                         self.script_buffer.get_end_iter(),
                                         True)
    script_scope = {}
    exec(script, script_scope)
    sequence = script_scope.get('sequence')
    sequence.connect('step-done', self._onSeqStepDone, button)
    if sequence and len(sequence.steps) > 0:
      button.set_sensitive(False)
      first_action = sequence.steps[0]
      self.progress_bar.set_fraction(0.0)
      self.progress_bar.set_text(str(first_action))
      sequence.startReally(False)

  def _onSeqStepDone(self, sequence, step, button):
    fraction = float(step + 1)/len(sequence.steps)
    self.progress_bar.set_fraction(fraction)
    if fraction >= 1.0:
      button.set_sensitive(True)
      self.progress_bar.set_text(_('Done'))
    else:
      action = sequence.steps[step + 1]
      self.progress_bar.set_text(str(action))

  def _onRecordChange(self, gobject, pspec, button):
    is_recording = self.script_buffer.get_property('recording')
    if is_recording:
      button.set_label(Gtk.STOCK_MEDIA_STOP)
    else:
      button.set_label(Gtk.STOCK_MEDIA_RECORD)

  def _onDelete(self, window, event):
    if self._askLoseChanges():
      self.script_buffer.clearBuffer()
      self.hide()
    return True

  def _onSave(self, button):
    '''
    Callback for 'save' button. Raises file chooser dialog for saving
    contents of script buffer.

    @param button: Button that was clicked.
    @type button: Gtk.Button
    '''
    save_dialog = Gtk.FileChooserDialog(
      'Save recorded script',
      action=Gtk.FileChooserAction.SAVE,
      buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
               Gtk.STOCK_OK, Gtk.ResponseType.OK))
    save_dialog.set_do_overwrite_confirmation(True)
    save_dialog.set_default_response(Gtk.ResponseType.OK)
    save_dialog.show_all()
    response = save_dialog.run()
    if response == Gtk.ResponseType.OK:
      save_to = open(save_dialog.get_filename(), 'w')
      save_to.write(self.script_buffer.get_text(
          self.script_buffer.get_start_iter(),
          self.script_buffer.get_end_iter(), True))
      save_to.close()
      self.script_buffer.set_modified(False)
    save_dialog.destroy()

  def _onClear(self, button):
    '''
    Callback for 'clear' button press.

    @param button: Button that was clicked.
    @type button: Gtk.Button
    '''
    if self._askLoseChanges():
      self.script_buffer.clearBuffer()

  def _onRecord(self, button):
    is_recording = self.script_buffer.get_property('recording')
    if not is_recording:
      self.script_buffer.startRecord()
    else:
      self.script_buffer.stopRecord()

  def _askLoseChanges(self):
    '''
    Raises a dialog that asks the user to confirm the loss of the current
    script in the buffer.

    @return: True if user confirms.
    @rtype: boolean
    '''
    if not self.script_buffer.get_modified():
      return True
    dialog = Gtk.MessageDialog(self.get_toplevel(), 0, Gtk.MessageType.WARNING,
                               Gtk.ButtonsType.OK_CANCEL,
                               _('The current script will be lost.'))
    dialog.set_title(_('Confirm clear'))
    response_id = dialog.run()
    dialog.destroy()
    if response_id == Gtk.ResponseType.OK:
      return True
    else:
      return False

class ScriptBuffer(GtkSource.Buffer):
  __gproperties__ = {
    'recording': (GObject.TYPE_BOOLEAN,
                  'Is recording',
                  'True if script buffer is recording',
                  False, GObject.PARAM_READWRITE)}
  factory_mapping = {'Level1' : script_factory.Level1SequenceFactory,
                     'Level2' : script_factory.Level2SequenceFactory}
  def __init__(self, uimanager):
    GtkSource.Buffer.__init__(self)
    lm = GtkSource.LanguageManager()
    lang = lm.guess_language(None, 'text/x-python')
    self.set_language(lang)
    self.set_highlight_syntax(True)
    self.script_factory = self.factory_mapping['Level2'](True)
    self._recording = False
    self._uimanager = uimanager
    self._addToUIManager()

  def _addToUIManager(self):
    self.script_type_actions = Gtk.ActionGroup('ScriptTypes')
    self.script_type_actions.add_radio_actions(
        (('Level1', None, 'Level 1', None, None, 1),
         ('Level2', None, 'Level 2', None, None, 2)),
        2, self._onChange)
    self._wait_for_focus_toggle = Gtk.ToggleAction('WaitForFocus',
                                                   'Record focus events',
                                                   None, None)
    self._wait_for_focus_toggle.set_active(True)
    self._wait_for_focus_toggle.connect('toggled', self._onWaitForFocusToggled)
    self.script_type_actions.add_action(self._wait_for_focus_toggle)
    script_type_ui = '''
<ui>
<popup>
    <placeholder name="ScriptType">
      <menuitem action="Level1" />
      <menuitem action="Level2" />
      <separator />
      <menuitem action="WaitForFocus" />
     </placeholder>
</popup>
</ui>'''
    self._uimanager.add_ui_from_string(script_type_ui)
    self._uimanager.insert_action_group(self.script_type_actions, 0)

  def startRecord(self):
    pyatspi.Registry.registerEventListener(self._onWindowActivate,
                                           'window:activate')
    pyatspi.Registry.registerEventListener(self._onFocus,
                                           'focus')
    pyatspi.Registry.registerEventListener(self._onDocLoad,
                                           'document:load-complete')
    pyatspi.Registry.registerKeystrokeListener(
      self._onKeystroke,
      mask = list(range(256)),
      kind=(pyatspi.KEY_PRESSED_EVENT, pyatspi.KEY_RELEASED_EVENT))
    self.set_property('recording', True)

  def stopRecord(self):
    pyatspi.Registry.deregisterEventListener(self._onWindowActivate,
                                             'window:activate')
    pyatspi.Registry.deregisterEventListener(self._onFocus,
                                             'focus')
    pyatspi.Registry.deregisterEventListener(self._onDocLoad,
                                             'document:load-complete')
    pyatspi.Registry.deregisterKeystrokeListener(
      self._onKeystroke,
      mask = list(range(256)),
      kind=(pyatspi.KEY_PRESSED_EVENT, pyatspi.KEY_RELEASED_EVENT))
    self.script_factory.terminateScript()
    while self.script_factory.commands_queue.qsize():
      self._appendText(self.script_factory.commands_queue.get_nowait())
    self.set_property('recording', False)

  def clearBuffer(self):
    '''
    Clears the script text buffer and inserts the intepreter and import line.
    '''
    self.set_text('%s\n\n%s\n\n' % \
                    (self.script_factory.intepreter_line,
                     self.script_factory.import_line))
    self.set_modified(False)

  def _onWindowActivate(self, event):
    '''
    Callback for window focus events. Calls the script factory and flushes
    it's queue.

    @param event: Focus at-spi event.
    @type event: Accessibility.Event
    '''
    if self._isMyApp(event.source):
      return
    self.script_factory.windowActivateCommand(event)

  def _onFocus(self, event):
    '''
    Callback for focus events. Calls the script factory and flushes
    it's queue.

    @param event: Focus at-spi event.
    @type event: Accessibility.Event
    '''
    if self._isMyApp(event.source):
      return
    app = event.source.getApplication()
    triggering_hotkey = None
    if getattr(app, 'name', None) == 'gnome-panel':
      if event.source.getRole() == pyatspi.ROLE_MENU and \
            event.source.parent.getRole() == pyatspi.ROLE_MENU_BAR:
        # A wild assumption that this was triggered with <Alt>F1
        triggering_hotkey = '<Alt>F1'
      if event.source.getRole() == pyatspi.ROLE_COMBO_BOX:
        triggering_hotkey = '<Alt>F2'
    elif getattr(app, 'name', None) == 'gnome-screenshot' and \
          event.source.getRole() == pyatspi.ROLE_TEXT and \
          pyatspi.getPath(event.source) == [0, 0, 0, 0, 1, 1]:
      triggering_hotkey = 'Print'

    if triggering_hotkey is not None:
      fake_event = _FakeDeviceEvent(triggering_hotkey, pyatspi.KEY_PRESSED_EVENT)
      self._onKeystroke(fake_event)

    self.script_factory.focusCommand(event)

    if triggering_hotkey is not None:
      fake_event = _FakeDeviceEvent(triggering_hotkey, pyatspi.KEY_RELEASED_EVENT)
      self._onKeystroke(fake_event)

    while self.script_factory.commands_queue.qsize():
      self._appendText(self.script_factory.commands_queue.get_nowait())

  def _onKeystroke(self, event):
    '''
    Callback for key press events. Calls the script factory and flushes
    it's queue.

    @param event: Key press at-spi event.
    @type event: Accessibility.DeviceEvent
    '''
    if event.type == pyatspi.KEY_PRESSED_EVENT:
      self.script_factory.keyPressCommand(event)
    elif event.type == pyatspi.KEY_RELEASED_EVENT:
      self.script_factory.keyReleaseCommand(event)
    while self.script_factory.commands_queue.qsize():
      self._appendText(self.script_factory.commands_queue.get_nowait())

  def _onDocLoad(self, event):
    self.script_factory.docLoadCommand()
    while self.script_factory.commands_queue.qsize():
      self._appendText(self.script_factory.commands_queue.get_nowait())

  def _isMyApp(self, acc):
    global APP_ID
    if APP_ID is not None:
      app = acc.getApplication()
      return getattr(app, 'name', None) == APP_ID
    else:
      return False

  def _appendText(self, text):
    self.insert(self.get_end_iter(), text)

  def do_set_property(self, pspec, value):
    if pspec.name == 'recording':
      self._recording = value

  def do_get_property(self, pspec):
    if pspec.name == 'recording':
      return self._recording

  def _onChange(self, action, current):
    factory = self.factory_mapping.get(current.get_name(),
                                       script_factory.Level1SequenceFactory)
    self.script_factory = factory(self._wait_for_focus_toggle.get_active())

  def _onWaitForFocusToggled(self, action):
    factory = self.script_factory.__class__
    self.script_factory = factory(self._wait_for_focus_toggle.get_active())

class _FakeDeviceEvent(object):
  def __init__(self, key_combo, type):
    id = Gdk.keyval_from_name(key_combo)
    if Gdk.keyval_from_name(key_combo):
      modifiers = 0
    else:
      id, modifiers = Gtk.accelerator_parse(key_combo)
    keymap = Gdk.Keymap.get_default()
    success, map_entry = keymap.get_entries_for_keyval(65471)
    self.type = type
    self.id = id
    self.hw_code = map_entry[0].keycode
    self.modifiers = int(modifiers)
    self.timestamp = 0
    self.event_string = Gdk.keyval_name(id)
    self.is_text = True



# Macaroon - a desktop macro tool
# Copyright (C) 2007 Eitan Isaacson <eitan@ascender.com>
# All rights reserved.

# This file may be distributed and/or modified under the terms of
# the GNU General Public License version 2 as published by
# the Free Software Foundation.
# This file is distributed without any warranty; without even the implied
# warranty of merchantability or fitness for a particular purpose.
# See "COPYING" in the source distribution for more information.

# Headers in this file shall remain intact.

import gi
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk
import pyatspi

keymap = Gdk.Keymap.get_default()

def getKeyCodeFromVal(keyval):
  global keymap
  success, entries = keymap.get_entries_for_keyval(keyval)
  entry = [(int(i.keycode), i.group, i.level) for i in entries]
  if not entry:
    raise TypeError("Invalid key name")
  else:
    return entry[0][0]

def getActiveFrame():
  desktop = pyatspi.Registry.getDesktop(0)
  active_frame = None
  for app in desktop:
    if app is None: continue
    for acc in app:
      if acc is None: continue
      if acc.getRole() == pyatspi.ROLE_FRAME:
        state_set = acc.getState()
        if state_set.contains(pyatspi.STATE_ACTIVE):
          active_frame = acc
        state_set.unref()
        if active_frame is not None:
          return active_frame
  return None



# Macaroon - a desktop macro tool
# Copyright (C) 2007 Eitan Isaacson <eitan@ascender.com>
# All rights reserved.

# This file may be distributed and/or modified under the terms of
# the GNU General Public License version 2 as published by
# the Free Software Foundation.
# This file is distributed without any warranty; without even the implied
# warranty of merchantability or fitness for a particular purpose.
# See "COPYING" in the source distribution for more information.

# Headers in this file shall remain intact.

from gi.repository import Gtk
from gi.repository import Gdk

import pyatspi
from queue import Queue
from time import time

class _CommandsQueue(Queue):
  def __init__(self):
    Queue.__init__(self)
    self.virgin = True
  def put(self, item, block=True, timeout=None):
    Queue.put(self, item, block, timeout)
    self.virgin = False

class ScriptFactory(object):
  '''
  Abstract class for a script factory. Classes of specific platforms like
  dogtail are derived from this.

  @cvar intepreter_line: The top intepreter line
  @type intepreter_line: string
  @cvar import_line: The import line for the specific platform
  @type import_line: strin
  @cvar MODIFIERS: Key symbols that are considered modifiers.
  @type MODIFIERS: list

  @ivar commands_queue: The commands that are being produced by the facory.
  @type commands_queue: Queue.Queue
  @ivar app_name: The last focused application name
  @type app_name: string
  @ivar frame_name: The last focused window frame name.
  @type frame_name: string.
  '''
  intepreter_line = '#!/usr/bin/python3.2'
  import_line = ''
  MODIFIERS = [Gdk.KEY_Control_L, Gdk.KEY_Control_R,
               Gdk.KEY_Alt_L, Gdk.KEY_Alt_R,
               Gdk.KEY_Super_L, Gdk.KEY_Super_R,
               Gdk.KEY_Shift_L, Gdk.KEY_Shift_R]

  def __init__(self):
    '''
    Initialize the script factory.
    '''
    self.commands_queue = _CommandsQueue()
    self.app_name = ''
    self.frame_name = ''

  def keyPressCommand(self, event):
    '''
    Processing key presses in to commands

    @param event: The keypress at-spi event.
    @type event: Accessibility.DeviceEvent
    '''
    pass

  def keyReleaseCommand(self, event):
    '''
    Processing key releases in to commands

    @param event: The key release at-spi event.
    @type event: Accessibility.DeviceEvent
    '''
    pass

  def windowActivateCommand(self, event):
    '''
    Store the focus event source's application name and frame name.

    @param event: The at-spi focus event.
    @type event: Accessibility.Event
    '''
    app = event.source.getApplication()
    if app:
      self.app_name = app.name
    self.frame_name = event.source.name

  def focusCommand(self):
    pass

  def docLoadCommand(self):
    pass

  def terminateScript(self):
    self.commands_queue.virgin = True

class SequenceFactory(ScriptFactory):
  import_line = \
      'from macaroon.playback import *\n\nsequence = MacroSequence()'
  def __init__(self, wait_for_focus=False):
    '''
    Initialize the object.
    '''
    ScriptFactory.__init__(self)
    self._time = 0
    if wait_for_focus:
      self._focus_comment = ''
    else:
      self._focus_comment = '#'
  def _getDelta(self):
    current_time = time()
    delta = current_time - self._time
    self._time = current_time
    if delta == current_time: delta = 0
    return int(delta*1000)
  def terminateScript(self):
    self.commands_queue.put_nowait('\nsequence.start()\n')
    ScriptFactory.terminateScript(self)

class Level2SequenceFactory(SequenceFactory):
  def __init__(self, wait_for_focus=False):
    SequenceFactory.__init__(self, wait_for_focus)
    self.typed_text = ''
    self.last_focused = None
    self.frame_name = ''
    self._last_app = None

  def keyPressCommand(self, event):
    if event.id in self.MODIFIERS or \
          (event.event_string.startswith('ISO') and \
             event.event_string != 'ISO_Left_Tab'):
      return
    if isinstance(event, pyatspi.deviceevent.DeviceEvent):
      # If it's a fake one, then it is a global WM hotkey, no need for context.
      self._prependContext()
    if event.modifiers in (0, Gdk.ModifierType.SHIFT_MASK) and \
          Gdk.keyval_to_unicode(event.id):
      self.typed_text += chr(Gdk.keyval_to_unicode(event.id))
    else:
      if self.frame_name:
        if isinstance(event, pyatspi.deviceevent.DeviceEvent):
          self.commands_queue.put_nowait(
            'sequence.append(WaitForWindowActivate("%s", None))\n' % \
              self.frame_name.replace('"','\"'))
        self.frame_name = ''
      if self.last_focused:
        name, path, role = self.last_focused
        self.commands_queue.put_nowait(
            '%ssequence.append(WaitForFocus("%s", acc_role=pyatspi.%s))\n' % \
            (self._focus_comment, name.replace('"','\"'), repr(role)))
        self.last_focused = None
      if self.typed_text:
        self.commands_queue.put_nowait(
          'sequence.append(TypeAction("%s"))\n' % \
            self.typed_text.replace('"','\"'))
        self.typed_text = ''
      self.commands_queue.put_nowait(
        'sequence.append(KeyComboAction("%s"))\n' % \
          Gtk.accelerator_name(event.id, Gdk.ModifierType(event.modifiers)))

  def focusCommand(self, event):
    try:
      path = pyatspi.getPath(event.source)
    except RuntimeError:
      path = []
    self.last_focused = (event.source.name,
                         path,
                         event.source.getRole())

  def windowActivateCommand(self, event):
    app = event.source.getApplication()
    if app is None:
      self.frame_name = event.source.name
      return
    if self._last_app == app.name:
      return
    else:
      self._last_app = app.name
    self.frame_name = event.source.name

  def docLoadCommand(self):
    print('factory thing')
    self.commands_queue.put_nowait(
      'sequence.append(WaitForDocLoad())\n')

  def _prependContext(self):
    if not self.frame_name and self.commands_queue.virgin:
      self.frame_name = self._getActiveFrameName()
    if self.frame_name:
      self.commands_queue.put_nowait(
        'sequence.append(WaitForWindowActivate("%s",None))\n' % \
          self.frame_name)
      self.frame_name = ''


  def _getActiveFrameName(self):
    desktop = pyatspi.Registry.getDesktop(0)
    active_frame = None
    for app in desktop:
      if not app: continue
      for acc in app:
        if acc is None: continue
        if acc.getRole() == pyatspi.ROLE_FRAME:
          state_set = acc.getState()
          if state_set.contains(pyatspi.STATE_ACTIVE):
            active_frame = acc
          state_set.unref()
          if active_frame is not None:
            return active_frame.name
    return ''

  def terminateScript(self):
    if self.typed_text:
      self._prependContext()
      self.commands_queue.put_nowait(
        'sequence.append(TypeAction           ("%s"))\n' % self.typed_text)
    SequenceFactory.terminateScript(self)

class Level1SequenceFactory(SequenceFactory):
  def keyPressCommand(self, event):
    delta = self._getDelta()
    self.commands_queue.put_nowait(
      'sequence.append(KeyPressAction(%d, %d, "%s")) # Press %s\n' % \
        (delta, event.hw_code, event.event_string, event.event_string))
  def keyReleaseCommand(self, event):
    delta = self._getDelta()
    self.commands_queue.put_nowait(
      'sequence.append(KeyReleaseAction(%d, %d, "%s")) # Release %s\n' % \
        (delta, event.hw_code, event.event_string, event.event_string))
  def windowActivateCommand(self, event):
    self.commands_queue.put_nowait(
      'sequence.append(WaitForWindowActivate("%s", None))\n' % \
        event.source.name)
  def focusCommand(self, event):
    self.commands_queue.put_nowait(
      '%ssequence.append(WaitForFocus("%s", acc_role=pyatspi.%s))\n' % \
        (self._focus_comment, event.source.name,
         event.source.getRole().value_name.replace("ATSPI_", "")))

class DogtailFactory(ScriptFactory):
  '''
  Script factory for dogtail scripts.

  @ivar typed_text: Plain text that has been typed so far.
  @type typed_text: string
  '''
  import_line = 'from dogtail.procedural import *'
  def __init__(self):
    '''
    Initialize the object.
    '''
    ScriptFactory.__init__(self)
    self.typed_text = ''

  def keyPressCommand(self, event):
    '''
    Create command lines for variuos key presses.

    @param event: Key press at-spi event.
    @type event: Accessibility.DeviceEvent
    '''
    if event.id in self.MODIFIERS or \
          event.event_string.startswith('ISO'):
      return
    if event.modifiers in (0, Gdk.ModifierType.SHIFT_MASK) and \
          Gdk.keyval_to_unicode(event.id):
      self.typed_text += chr(Gdk.keyval_to_unicode(event.id))
    else:
      if self.app_name:
        self.commands_queue.put_nowait('focus.application("%s")\n' % \
                                         self.app_name)
        self.app_name = ''
      if self.frame_name:
        self.commands_queue.put_nowait('focus.frame("%s")\n' % \
                                         self.frame_name)
        self.frame_name = ''
      if self.typed_text:
        self.commands_queue.put_nowait('type("%s")\n' % \
                                         self.typed_text)
        self.typed_text = ''
      self.commands_queue.put_nowait('keyCombo("%s")\n' % \
                                       Gtk.accelerator_name(event.id,
                                                            Gdk.ModifierType(event.modifiers)))


class NativeFactory(DogtailFactory):
  '''
  Script factory for native scripts. Besides the import line,
  it should be identical to dogtail scripts.
  '''
  import_line = 'from accerciser.script_playback import *'

class LDTPFactory(DogtailFactory):
  '''
  Script factory for LDTP scripts.

  @ivar typed_text: Plain text that has been typed so far.
  @type typed_text: string
  '''
  import_line = 'from ldtp import *'
  def __init__(self):
    '''
    Initialize the object.
    '''
    ScriptFactory.__init__(self)
    self.typed_text = ''

  def keyPressCommand(self, event):
    '''
    Create command lines for variuos key presses.

    @param event: Key press at-spi event.
    @type event: Accessibility.DeviceEvent
    '''
    if event.id in self.MODIFIERS or \
          event.event_string.startswith('ISO'):
      return
    if event.modifiers in (0, Gdk.ModifierType.SHIFT_MASK) and \
          Gdk.keyval_to_unicode(event.id):
      self.typed_text += chr(Gdk.keyval_to_unicode(event.id))
    else:
      if self.frame_name:
        self.commands_queue.put_nowait('waittillguiexist("%s")\n' % \
                                         self.frame_name)
        self.frame_name = ''
      if self.typed_text:
        self.commands_queue.put_nowait('generatekeyevent("%s")\n' % \
                                         self.typed_text)
        self.typed_text = ''
      self.commands_queue.put_nowait('generatekeyevent("%s")\n' % \
                                       Gtk.accelerator_name(event.id,
                                                            Gdk.ModifierType(event.modifiers)))



'''
Keypress actions.

@author: Eitan Isaacson
@copyright: Copyright (c) 2007 Eitan Isaacson
@license: GPL

This file may be distributed and/or modified under the terms of the GNU General
Public License version 2 as published by the Free Software Foundation. This file
is distributed without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose.

See "COPYING" in the source distribution for more information.

Headers in this file shall remain intact.
'''

from gi.repository import GObject
from gi.repository import GLib
import pyatspi
import sys
from .wait_actions import WaitAction
from os import environ
_ = lambda x: x

class MacroSequence(GObject.GObject):
  '''
  Sequence class. Holds a list of steps and performs them in order.
  Waits for each step to emit a "done" signal. And performs the next step
  only after it's delta time has lapsed. Emits "step-done" each time a step
  is completed.

  @ivar _loop: Loop instance if the main loop should be embedded.
  @type _loop: GLib.MainLoop
  @ivar _verbose: Print to standard output the current step the sequence is in.
  @type _verbose: boolean
  @ivar _current_step: The index of the currently performed step.
  @type _current_step: integer
  @ivar _current_handler: Event handler ID of currently performed step.
  @type _current_handler: integer
  @ivar steps: List of sequence steps to perform.
  @type steps: list of L{SequenceStep}
  @ivar _anticipated_event_types: List of event types we should be on the lookout
  for.
  @type _anticipated_event_types: list of string
  @ivar _anticipated_events: Events that we have anticipated, and that have
  came through.
  @type _anticipated_events: list of Accessibility.Accessible
  '''
  __gsignals__ = {'step-done' : (GObject.SignalFlags.RUN_FIRST,
                                 None, (GObject.TYPE_INT,))}
  def __init__(self):
    '''
    Initialize L{MacroSequence}.
    '''
    super(MacroSequence, self).__init__()
    self._loop = None
    self._verbose = False
    self._current_step = 0
    self._current_handler = 0
    self.steps = []
    self._anticipated_event_types = []
    self._anticipated_events = []

  def append(self, step):
    '''
    Add a new sequence step to the end of the sequence.

    @param step: A sequence step to add.
    @type step: L{SequenceStep}
    '''
    self.steps.append(step)

  def start(self, embedded_loop=True, verbose=False):
    '''
    Start sequence.

    @param embedded_loop: Embed a loop in the sequnce, if we are running this
    outside of a main loop.
    @type embedded_loop: boolean
    '''
    self._verbose = bool(verbose or environ.get('MACAROON_VERBOSE', 0))
    self._iterAction()
    if embedded_loop:
      self._loop = GLib.MainLoop()
      self._loop.run()

  def _iterAction(self):
    '''
    Iterate to the next sequence step.
    '''
    if len(self.steps) <= self._current_step:
      if self._loop is not None:
        self._loop.quit()
      return
    action = self.steps[self._current_step]
    if self._verbose:
      print(_('SEQUENCE: %s') % action)

    try:
      next_action = self.steps[self._current_step + 1]
    except IndexError:
      next_action = None

    pyatspi.Registry.deregisterEventListener(self._onAnticipatedEvent,
                                             *self._anticipated_event_types)
    if isinstance(next_action, WaitAction):
      self._anticipated_event_types = next_action.wait_for
    else:
      self._anticipated_event_types = []
    pyatspi.Registry.registerEventListener(self._onAnticipatedEvent,
                                           *self._anticipated_event_types)
    self._current_handler = action.connect('done', self._onStepDone)

    GLib.timeout_add(action.delta_time, self._doAction, action)

  def _onAnticipatedEvent(self, event):
    '''
    Callback for antibitaed events, keep them for the next sequence step which
    might be a L{WaitAction} .

    @param event: Event that was recieved.
    @type event: Accessibility.Event
    '''
    self._anticipated_events.append(event)

  def _onStepDone(self, action):
    '''
    Callback for when a step is done.

    @param action: Step that has finished.
    @type action: L{SequenceStep}
    '''
    action.disconnect(self._current_handler)
    self.emit('step-done', self._current_step)
    self._current_step += 1
    self._iterAction()

  def _doAction(self, action):
    '''
    Perform the action.

    @param action: Step to perform.
    @type action: L{SequenceStep}
    '''
    if isinstance(action, WaitAction):
      action(self._anticipated_events)
      self._anticipated_events = []
    else:
      action()
    return False



'''
Defines the manager for global hot keys.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository.Gio import Settings as GSettings

from .i18n import _
import pyatspi

HOTKEYS_GSCHEMA = 'org.a11y.Accerciser.hotkeys'
HOTKEYS_BASEPATH = '/org/a11y/accerciser/hotkeys/'

COL_COMPONENT = 0
COL_DESC = 1
COL_CALLBACK = 2
COL_KEYPRESS = 3
COL_MOD = 4
COL_LOCALIZED_COMP = 5

def _charToKeySym(key):
  '''
  A convinience function to convert either a character, or key name to it's
  respective keyval

  @param key: The character or key name to convert.
  @type key: string

  @return: A key symbol
  @rtype: long
  '''
  try:
    rv = gdk.unicode_to_keyval(ord(key))
  except:
    rv = getattr(gdk, 'KEY_%s' % key)
  return rv

class HotkeyManager(gtk.ListStore):
  '''
  A model that stores all of the global key bindings. All accerciser components
  that need global hotkeys should register the key combination and callback
  with the main instance of this class.
  '''
  def __init__(self):
    '''
    Constructor for the L{HotkeyManager}
    '''
    gtk.ListStore.__init__(self, str, str, object, int, int, str)
    self.connect('row-changed', self._onComboChanged)

    masks = [mask for mask in pyatspi.allModifiers()]
    pyatspi.Registry.registerKeystrokeListener(
      self._accEventKeyPressed, mask=masks, kind=(pyatspi.KEY_PRESSED_EVENT,))


  def _accEventKeyPressed(self, event):
    '''
    Handle certain key presses globally. Pass on to the hotkey manager the
    key combinations pressed for further processing.

    @param event: The event that is being handled.
    @type event: L{pyatspi.event.Event}
    '''
    handled = self.hotkeyPress(event.hw_code, event.modifiers)
    event.consume = handled


  def hotkeyPress(self, key, modifiers):
    '''
    Call the appropriate callbacks for given key combination. This method
    should be called by an at-spi keyboard:press event handler in the
    main program.

    @param key: The pressed key code.
    @type key: integer
    @param modifiers: The modifiers that were depressed during the keystroke.
    @type modifiers: integer
    '''
    km = gdk.Keymap.get_default()

    callback = None

    for combo in self:
      success, entries = km.get_entries_for_keyval(combo[COL_KEYPRESS])
      if not success: continue
      if key in [int(entry.keycode) for entry in entries] and \
            modifiers & combo[COL_MOD] == combo[COL_MOD]:
        callback = combo[COL_CALLBACK]
        if callback:
          callback()
    return bool(callback)

  def addKeyCombo(self, component, localized_component, description,
                  callback, keypress, modifiers):
    '''
    Adds the given key combination with the appropriate callbacks to
    the L{HotkeyManager}. If an identical description with the identical
    component already exists in the model, just reassign with the new callback.

    I{Note:} It is important that the component and description strings be
    unique.

    @param component: The component name, usually the plugin name, or "Core".
    @type component: string
    @param description: A description of the action performed during the given
    keycombo.
    @type description: string
    @param callback: The callback to call when the given key combination
    is pressed.
    @type callback: callable
    @param keypress: The key symbol of the keystroke that performs given operation.
    @type keypress: long
    @param modifiers: The modifiers that must be depressed for function to
    be perfomed.
    @type modifiers: int
    '''
    component_desc_pairs = list(zip([row[COL_COMPONENT] for row in self],
                               [row[COL_DESC] for row in self]))
    if (component, description) in component_desc_pairs:
      path = component_desc_pairs.index((component, description))
      self[path][COL_CALLBACK] = callback
    else:
      gspath = self._getComboGSettingsPath(component, description)
      gsettings = GSettings.new_with_path(HOTKEYS_GSCHEMA, gspath)
      if gsettings.get_string('hotkey-combo'):
        final_keypress, final_modifiers = gtk.accelerator_parse(
          gsettings.get_string('hotkey-combo'))
      else:
        final_keypress, final_modifiers = keypress, modifiers
      self.append([component, description, callback,
                   int(final_keypress), final_modifiers, localized_component])

  def removeKeyCombo(self, component, description, callback, key, modifiers):
    '''
    Removes the given callback from L{HotkeyManager}. It does not erase the
    entire key combo entry.

    @param component: The component name, usually the plugin name, or "Core".
    @type component: string
    @param description: A description of the action performed during the given
    keycombo.
    @type description: string
    @param callback: The callback to call when the given key combination
    is pressed.
    @type callback: callable
    @param key: The key symbol of the keystroke that performs given operation.
    @type key: long
    @param modifiers: The modifiers that must be depressed for function to
    be perfomed.
    @type modifiers: int
    '''
    iter = self.get_iter_first()
    while iter:
      if self[iter][COL_CALLBACK] == callback:
        # We never really remove it, just set the callback to None
        self[iter][COL_CALLBACK] = ''
      iter = self.iter_next(iter)

  def _onComboChanged(self, model, path, iter):
    '''
    Callback for row changes. Copies the changed key combos over to gsettings.

    @param model: The model that emitted the signal. Should be this class instance.
    @type model: L{gtk.TreeModel}
    @param path: The path of the row that has changed.
    @type path: tuple
    @param iter: The iter of the row that has changed.
    @type iter: L{gtk.TreeIter}
    '''
    if not model[iter][COL_COMPONENT] or not model[iter][COL_DESC]:
      return

    gspath = self._getComboGSettingsPath(model[iter][COL_COMPONENT],
                                         model[iter][COL_DESC])
    gsettings = GSettings.new_with_path(HOTKEYS_GSCHEMA, gspath)
    combo_name = gtk.accelerator_name(model[iter][COL_KEYPRESS],
                                      gdk.ModifierType(model[iter][COL_MOD]))

    key = gsettings.get_string('hotkey-combo')

    if key != combo_name and key != '/':
      gsettings.set_string('hotkey-combo', combo_name)


  def _getComboGSettingsPath(self, component, description):
    '''
    Useful method that build and returns a gsettings path for a key combo.

    @param component: The component of the hotkey.
    @type component: string
    @param description: The description of the hotkey action
    @type description: string

    @return: A full gsettings path
    @rtype: string
    '''
    dash_component = self.__dasherize(component)
    dash_description = self.__dasherize(description)

    path = '/'.join([dash_component, dash_description])

    return HOTKEYS_BASEPATH + path + '/'


  def __dasherize(self, item):
    '''
    This method dasherize and decapitalize a given string.

    @param component: The given string
    @type component: string

    @return: A dasherized and decapitalized string
    @rtype: string
    '''
    return item.lower().replace(' ', '-')

class HotkeyTreeView(gtk.TreeView):
  '''
  A tree view of the variuos global hotkey combinations. The keys and
  modifiers could also be changed through this widget.
  '''
  def __init__(self, hotkey_manager):
    '''
    Construct the tree view with the given L{HotkeyManager}.

    @ivar hotkey_manager: The manager we wish to view.
    @type hotkey_manager: L{HotkeyManager}

    @param hotkey_manager: The manager we wish to view.
    @type hotkey_manager: L{HotkeyManager}
    '''
    gtk.TreeView.__init__(self)
    self.hotkey_manager = hotkey_manager
    modelfilter = self.hotkey_manager.filter_new(None)
    modelfilter.set_visible_func(self._rowVisibleFunc, None)
    self.set_model(modelfilter)
    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Component'))
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', COL_COMPONENT)
    tvc.set_cell_data_func(crt, self._componentDataFunc, COL_COMPONENT)
    self.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Task'))
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', COL_DESC)
    tvc.set_cell_data_func(crt, self._translateDataFunc, COL_DESC)
    self.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Key'))
    tvc.set_min_width(64)
    tvc.pack_start(crt, True)
    crt.props.editable = True
    tvc.add_attribute(crt, 'text', COL_KEYPRESS)
    tvc.set_cell_data_func(crt, self._keyCellFunc)
    crt.connect('edited', self._onKeyChanged)
    self.append_column(tvc)

    crt = gtk.CellRendererToggle()
    tvc = gtk.TreeViewColumn(_('Alt'))
    tvc.pack_start(crt, True)
    tvc.set_cell_data_func(crt, self._modCellFunc, gdk.ModifierType.MOD1_MASK)
    crt.connect('toggled', self._onModToggled, gdk.ModifierType.MOD1_MASK)
    self.append_column(tvc)

    crt = gtk.CellRendererToggle()
    tvc = gtk.TreeViewColumn(_('Ctrl'))
    tvc.pack_start(crt, True)
    tvc.set_cell_data_func(crt, self._modCellFunc, \
                           gdk.ModifierType.CONTROL_MASK)
    crt.connect('toggled', self._onModToggled, gdk.ModifierType.CONTROL_MASK)
    self.append_column(tvc)

    crt = gtk.CellRendererToggle()
    tvc = gtk.TreeViewColumn(_('Shift'))
    tvc.pack_start(crt, True)
    tvc.set_cell_data_func(crt, self._modCellFunc, gdk.ModifierType.SHIFT_MASK)
    crt.connect('toggled', self._onModToggled, gdk.ModifierType.SHIFT_MASK)
    self.append_column(tvc)

  def _translateDataFunc(self, column, cell, model, iter, column_id):
    '''
    Show the component name as a translated string.

    @param column: The treeview column of the cell renderer.
    @type column: L{gtk.TreeViewColumn}
    @param cell: The cell rendere we need to modify.
    @type cell: L{gtk.CellRendererText}
    @param model: The treeview's model.
    @type model: L{gtk.ListStore}
    @param iter: The iter of the given cell data.
    @type iter: L{gtk.TreeIter}
    '''
    cell.set_property('text', _(model[iter][column_id]))

  def _componentDataFunc(self, column, cell, model, iter, column_id):
    '''
    Show the component name as a translated string.

    @param column: The treeview column of the cell renderer.
    @type column: L{gtk.TreeViewColumn}
    @param cell: The cell rendere we need to modify.
    @type cell: L{gtk.CellRendererText}
    @param model: The treeview's model.
    @type model: L{gtk.ListStore}
    @param iter: The iter of the given cell data.
    @type iter: L{gtk.TreeIter}
    '''
    cell.set_property('text', model[iter][COL_LOCALIZED_COMP] or \
                        model[iter][COL_COMPONENT])

  def _keyCellFunc(self, column, cell, model, iter, foo=None):
    '''
    Show the key symbol as a string for easy readability.

    @param column: The treeview column of the cell renderer.
    @type column: L{gtk.TreeViewColumn}
    @param column: The cell rendere we need to modify.
    @type column: L{gtk.CellRendererText}
    @param model: The treeview's model.
    @type model: L{gtk.ListStore}
    @param iter: The iter of the given cell data.
    @type iter: L{gtk.TreeIter}
    '''
    if model[iter][COL_KEYPRESS] > 0:
      cell.set_property('text',
                        gdk.keyval_name(model[iter][COL_KEYPRESS]))
      cell.set_property('sensitive', True)
    else:
      cell.set_property('text', '<select key>')
      cell.set_property('sensitive', False)

  def _modCellFunc(self, column, cell, model, iter, mask):
    '''
    Show the given modifier mask as toggled or not.

    @param column: The treeview column of the cell renderer.
    @type column: L{gtk.TreeViewColumn}
    @param column: The cell rendere we need to modify.
    @type column: L{gtk.CellRendererText}
    @param model: The treeview's model.
    @type model: L{gtk.ListStore}
    @param iter: The iter of the given cell data.
    @type iter: L{gtk.TreeIter}
    @param mask: A modifier mask.
    @type mask: integer
    '''
    cell.set_property('active', bool(mask & model[iter][COL_MOD]))

  def _onKeyChanged(self, cellrenderertext, path, new_text):
    '''
    A callback for the key cellrenderer when 'edited'. Model must be
    changed accordingly.

    @param cellrenderertext: The cell renderer that emitted the signal
    @type cellrenderertext: L{gtk.CellRendererText}
    @param path: Path of the edited cellrenderer.
    @type path: tuple
    @param new_text: The new text that was entered.
    @type new_text: string
    '''
    keysym = -1
    if new_text:
      try:
        keysym = _charToKeySym(new_text)
      except:
        keysym = _charToKeySym(new_text[0])
    self.hotkey_manager[path][COL_KEYPRESS] = int(keysym)

  def _onModToggled(self, renderer_toggle, path, mask):
    '''
    A callback for the modifiers' cellrenderers when 'toggled'.
    Model must be changed accordingly.

    @param renderer_toggle: The cell renderer that emitted the signal
    @type renderer_toggle: L{gtk.CellRendererToggle}
    @param path: Path of the edited cellrenderer.
    @type path: tuple
    @param mask: Modifier mask that must be inverted.
    @type new_text: integer
    '''
    self.hotkey_manager[path][COL_MOD] ^= mask

  def _rowVisibleFunc(self, model, iter, foo=None):
    '''
    A filter function to hide the rows that do not contain valid callbacks.
    This is usually the case when a plugin is disabled.

    @param model: The view's model.
    @type model: L{gtk.ListStore}
    @param iter: The iter of the row in question.
    @type iter: L{gtk.TreeIter}

    @return: True if row should be displayed.
    @rtype: boolean
    '''
    return bool(model[iter][COL_CALLBACK])



'''
Keypress actions.

@author: Eitan Isaacson
@copyright: Copyright (c) 2007 Eitan Isaacson
@license: GPL

This file may be distributed and/or modified under the terms of the GNU General
Public License version 2 as published by the Free Software Foundation. This file
is distributed without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose.

See "COPYING" in the source distribution for more information.

Headers in this file shall remain intact.
'''

import gi
gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib
import pyatspi

from .sequence_step import AtomicAction
from . import utils

_ = lambda x: x

# Highest granularity, define timing for every single press and release

# Minimum delta time
min_delta = 50

# Maximum time before a key release
release_max = 400

keymap = Gdk.Keymap.get_default()

class KeyPressAction(AtomicAction):
  '''
  A key press step. Emulates a keyboard key press.
  '''
  def __init__(self, delta_time, key_code=None, key_name=None):
    '''
    Initialize L{KeyPressAction}. Could use either a hardware keycode,
    a key name, or both.

    @param delta_time: Time to wait before performing this step.
    @type delta_time: integer
    @param key_code: Hardware keycode.
    @type key_code: integer.
    @param key_name: Key name.
    @type key_name: string.
    '''
    if (None, None) == (key_name, key_code):
      raise TypeError("Need either a key code or a key name")
    if delta_time > release_max: delta_time = release_max
    self._key_name = key_name
    if key_code is None:
      key_code = utils.getKeyCodeFromVal(Gdk.keyval_from_name(key_name))
    AtomicAction.__init__(self, delta_time, self._keyPress, key_code)

  def _keyPress(self, key_code):
    '''
    Perform actual key press.

    @param key_code: Hardware key code.
    @type key_code: integer
    '''
    pyatspi.Registry.generateKeyboardEvent(key_code, None, pyatspi.KEY_PRESS)

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return _('Key press %s') % self._key_name or 'a key'

class KeyReleaseAction(AtomicAction):
  '''
  A key release step. Emulates a keyboard key release.
  '''
  def __init__(self, delta_time, key_code=None, key_name=None):
    '''
    Initialize L{KeyReleaseAction}. Could use either a hardware keycode,
    a key name, or both.

    @param delta_time: Time to wait before performing this step.
    @type delta_time: integer
    @param key_code: Hardware keycode.
    @type key_code: integer.
    @param key_name: Key name.
    @type key_name: string.
    '''
    if (None, None) == (key_name, key_code):
      raise TypeError("Need either a key code or a key name")
    if delta_time > release_max: delta_time = release_max
    self._key_name = key_name
    if key_code is None:
      key_code = utils.getKeyCodeFromVal(Gdk.keyval_from_name(key_name))
    AtomicAction.__init__(self, delta_time, self._keyRelease, key_code)

  def _keyRelease(self, key_code):
    '''
    Perform actual key release.

    @param key_code: Hardware key code.
    @type key_code: integer
    '''
    pyatspi.Registry.generateKeyboardEvent(key_code, None, pyatspi.KEY_RELEASE)

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return _('Key release %s') % self._key_name or 'a key'


# A bit smarter about common interactions.

keystroke_interval = 200
mod_key_code_mappings = {
  'GDK_CONTROL_MASK' : 37,
  'GDK_MOD1_MASK' : 64,
  'GDK_LOCK_MASK' : 66,
  'GDK_SHIFT_MASK' : 50
  }

class KeyComboAction(AtomicAction):
  '''
  Key combo action. Performs a press and release of a single or key combination.

  @ivar _key_combo: Name of key combination or single key press-release.
  @type _key_combo: string
  '''
  def __init__(self, key_combo, delta_time=0):
    '''
    Initialize L{KeyComboAction}.

    @param key_combo: Name of key combination or single key press-release.
    @type key_combo: string
    @param delta_time: Time to wait before performing step.
    @type delta_time: integer
    '''
    keyval, modifiers = Gtk.accelerator_parse(key_combo)
    key_code = utils.getKeyCodeFromVal(keyval)
    self._key_combo = key_combo
    if delta_time < min_delta: delta_time = min_delta
    AtomicAction.__init__(self, delta_time, self._doCombo, key_code, modifiers)

  def __call__(self):
    '''
    Perform step. Overridden to omit L{SequenceStep.stepDone}.
    '''
    self._func(*self._args)

  def _doCombo(self, key_code, modifiers):
    '''
    Perform combo operation.

    @param key_code: Key code to press.
    @type key_code: integer
    @param modifiers: Modifier mask to press with.
    @type modifiers: integer
    '''
    interval = 0
    mod_hw_codes = list(map(mod_key_code_mappings.get, modifiers.value_names))
    for mod_hw_code in mod_hw_codes:
      GLib.timeout_add(interval, self._keyPress, mod_hw_code)
      interval += keystroke_interval
    GLib.timeout_add(interval, self._keyPressRelease, key_code)
    interval += keystroke_interval
    mod_hw_codes.reverse()
    for mod_hw_code in mod_hw_codes:
      GLib.timeout_add(interval, self._keyRelease, mod_hw_code)
      interval += keystroke_interval
    GLib.timeout_add(interval, self.stepDone)

  def _keyPress(self, hw_code):
    '''
    Press modifier key.

    @param hw_code: Hardware code for key.
    @type hw_code: integer
    '''
    pyatspi.Registry.generateKeyboardEvent(hw_code, None, pyatspi.KEY_PRESS)
    return False

  def _keyRelease(self, hw_code):
    '''
    Release modifier key.

    @param hw_code: Hardware code for key.
    @type hw_code: integer
    '''
    pyatspi.Registry.generateKeyboardEvent(hw_code, None, pyatspi.KEY_RELEASE)
    return False

  def _keyPressRelease(self, keyval):
    '''
    Press and release non-modifier key.

    @param key_code: Key code.
    @type key_code: integer
    '''
    pyatspi.Registry.generateKeyboardEvent(keyval, None,
                                           pyatspi.KEY_PRESSRELEASE)
    return False

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return _('Press %s') % self._key_combo

class TypeAction(AtomicAction):
  '''
  Type a sequence of characters.
  '''
  def __init__(self, string_to_type, delta_time=0, interval=None):
    '''
    Initialize L{TypeAction}

    @param string_to_type: String to type.
    @type string_to_type: string
    @param delta_time: time before starting this step.
    @type delta_time: integer
    @param interval: Time between keystrokes.
    @type interval: integer
    '''
    self._string_to_type = string_to_type
    if interval:
      self.interval = interval
    else:
      self.interval = keystroke_interval
    if delta_time < min_delta: delta_time = min_delta
    AtomicAction.__init__(self, delta_time, self._doType, string_to_type)

  def __call__(self):
    '''
    Perform step. Overridden to omit L{SequenceStep.stepDone}.
    '''
    self._func(*self._args)

  def _doType(self, string_to_type):
    '''
    Do typing action.

    @param string_to_type: String to type.
    @type string_to_type: string
    '''
    interval = 0
    for char in string_to_type:
      GLib.timeout_add(interval, self._charType,
                          Gdk.unicode_to_keyval(ord(char)))
      interval += self.interval
    GLib.timeout_add(interval, self.stepDone)

  def _charType(self, keyval):
    '''
    Type a single character.

    @param keyval: Key code to type.
    @type keyval: intger
    '''
    success, entries = keymap.get_entries_for_keyval(keyval)
    entry = [(int(i.keycode), i.group, i.level) for i in entries]
    key_code, group, level = entry[0]
    #key_code, group, level = utils.keymap.get_entries_for_keyval(keyval)[0]
    if level == 1:
        pyatspi.Registry.generateKeyboardEvent(50, None,
                                               pyatspi.KEY_PRESS)
    pyatspi.Registry.generateKeyboardEvent(key_code, None,
                                           pyatspi.KEY_PRESSRELEASE)
    if level == 1:
        pyatspi.Registry.generateKeyboardEvent(50, None,
                                               pyatspi.KEY_RELEASE)
    return False

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return _('Type %s') % self._string_to_type



'''
Defines the menus used in the application.

@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''

from gi.repository import Gio as gio

from .i18n import _

main_menu = gio.Menu()
file_menu = gio.Menu()
main_menu.append_submenu(_('_File'), file_menu)
edit_menu = gio.Menu()
main_menu.append_submenu(_('_Edit'), edit_menu)
bookmarks_menu = gio.Menu()
bookmarks_menu_general_section = gio.Menu()
bookmarks_menu.append_section(None, bookmarks_menu_general_section)
bookmarks_menu_individual_section = gio.Menu()
bookmarks_menu.append_section(None, bookmarks_menu_individual_section)
main_menu.append_submenu(_('_Bookmarks'), bookmarks_menu)
view_menu = gio.Menu()
main_menu.append_submenu(_('_View'), view_menu)
view_menu_general_section = gio.Menu()
view_menu.append_section(None, view_menu_general_section)
view_menu_treeview_section = gio.Menu()
view_menu.append_section(None, view_menu_treeview_section)
help_menu = gio.Menu()
main_menu.append_submenu(_('_Help'), help_menu)

treeview_context_menu = gio.Menu()



'''
Checks if all required Python modules are installed.

@author: Peter Parente
@organization: IBM Corporation
@copyright: Copyright (c) 2005, 2007 IBM Corporation
@license: The BSD License

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD License which accompanies
this distribution, and is available at
U{http://www.opensource.org/licenses/bsd-license.php}
'''
import sys, os

PYGTK_REQ = '2.0'
PYATSPI_REQ = (2, 23, 3)
GTK_VERSION = (2, 8, 0)

try:
  # stop the gail bridge from loading during build
  val = os.environ['GTK_MODULES']
  os.environ['GTK_MODULES'] = val.replace('gail:atk-bridge', '')
except KeyError:
  pass

# test for python modules
modules = ['pyatspi', 'cairo', 'rsvg', 'gi', 'gi.repository.Gtk', \
          'gi.repository.GConf', 'gi.repository.Gdk', 'gi.repository.Atk', \
          'gi.repository.GObject', 'gi.repository.GdkPixbuf', \
          'gi.repository.Wnck', 'gi.repository.Gio']

for name in modules:
  try:
    m = __import__(name)
    print(name)
  except ImportError as e:
    print(name, '*MISSING*')
    sys.exit(1)
  except RuntimeError:
    # ignore other errors which might be from lack of a display
    continue
  if name =='pyatspi':
    try:
      compared = [cmp(*x) for x in zip(PYATSPI_REQ, m.__version__)]
    except AttributeError:
      # Installed pyatspi does not support __version__, too old.
      compared = [-1, 0, 0]
    if -1 in compared and 1 not in compared[:compared.index(-1)]:
      # A -1 without a 1 preceding it means an older version.
      print("\nNeed pyatspi 1.23.4 or higher (or SVN trunk)")
      sys.exit(1)
print()



'''
Checks if all required Python modules are installed.

@author: Peter Parente
@organization: IBM Corporation
@copyright: Copyright (c) 2005, 2007 IBM Corporation
@license: The BSD License

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD License which accompanies
this distribution, and is available at
U{http://www.opensource.org/licenses/bsd-license.php}
'''
import sys, os

PYGTK_REQ = '2.0'
GTK_VERSION = (2, 8, 0)

try:
  # stop the gail bridge from loading during build
  val = os.environ['GTK_MODULES']
  os.environ['GTK_MODULES'] = val.replace('gail:atk-bridge', '')
except KeyError:
  pass

# test for python modules
modules = ['pygtk', 'gtk', 'Gtk.gdk', 'wnck']
for name in modules:
  try:
    m = __import__(name)
    print(name)
  except ImportError as e:
    if name == 'wnck' and e.args[0].find('gtk') > -1:
      # just no display, continue
      continue
    print(name, '*MISSING*')
    sys.exit(1)
  except RuntimeError:
    # ignore other errors which might be from lack of a display
    continue
  if name == 'pygtk':
    m.require('2.0')
  elif name == 'gtk':
    m.check_version(*GTK_VERSION)

sys.exit(0)



# -*- coding: utf-8 -*-
# Macaroon - a desktop macro tool
# Copyright (C) 2007 Eitan Isaacson <eitan@ascender.com>
# All rights reserved.

# This file may be distributed and/or modified under the terms of
# the GNU General Public License version 2 as published by
# the Free Software Foundation.
# This file is distributed without any warranty; without even the implied
# warranty of merchantability or fitness for a particular purpose.
# See "COPYING" in the source distribution for more information.

# Headers in this file shall remain intact.

from gi.repository import Gtk, GObject
#from gnome import program_get, url_show

_ = lambda x: x
#Gtk.about_dialog_set_url_hook(lambda dialog, url, data: url_show(url), None)


class MacaroonAboutDialog(Gtk.AboutDialog):
  '''
  Creates a dialog with info about the program.

  @cvar AUTHORS: List of authors.
  @type AUTHORS: list of string
  @cvar ARTISTS: List of artists.
  @type ARTISTS:list of string
  @cvar DOCUMENTERS: List of documenters.
  @type DOCUMENTERS: list of string
  @cvar TRANSLATORS: Translator.
  @type TRANSLATORS: string
  @cvar COMMENTS: Comments about program.
  @type COMMENTS: string
  @cvar COPYRIGHT: Copyright notice.
  @type COPYRIGHT: string
  @cvar LICENSE: License of program.
  @type LICENSE: string
  @cvar WEB_URL: URL to program's website.
  @type WEB_URL: string
  @cvar WEB_LABEL: Label of URL button.
  @type WEB_LABEL: string
  '''
  AUTHORS = ['Eitan Isaacson <eitan@ascender.com>']
  ARTISTS = []
  DOCUMENTERS = []
  TRANSLATORS = _('translator-credits')
  COMMENTS = _('A desktop macro tool')
  COPYRIGHT =  _('macaroon Copyright © 2007 Eitan Isaacson')
  LICENSE = ''
  WEB_URL = 'http://live.gnome.org/Macaroon'
  WEB_LABEL = _('Web site')
  def __init__(self):
    '''
    Initialize dialog.
    '''
#    program = program_get()
    GObject.GObject.__init__(self)
    self.connect('response', self._onResponse)
    Gtk.AboutDialog.set_authors(self, self.AUTHORS)
    Gtk.AboutDialog.set_artists(self, self.ARTISTS)
    Gtk.AboutDialog.set_documenters(self, self.DOCUMENTERS)
    Gtk.AboutDialog.set_comments(self, self.COMMENTS)
    Gtk.AboutDialog.set_copyright(self, self.COPYRIGHT)
    Gtk.AboutDialog.set_license(self, self.LICENSE)
    Gtk.AboutDialog.set_logo_icon_name(self, 'macaroon')
#    Gtk.AboutDialog.set_version(self, program.get_app_version())
    Gtk.AboutDialog.set_website(self, self.WEB_URL)
    Gtk.AboutDialog.set_website_label(self, self.WEB_LABEL)

  def _onResponse(self, dialog, response_id):
    '''
    Callback for dialog responses, always destroy it.

    @param dialog: This dialog.
    @type dialog: L{MacaroonAboutDialog}
    @param response_id: Response ID recieved.
    @type response_id: integer
    '''
    self.destroy()



'''
Base sequence step class and some derivatives.

@author: Eitan Isaacson
@copyright: Copyright (c) 2007 Eitan Isaacson
@license: GPL

This file may be distributed and/or modified under the terms of the GNU General
Public License version 2 as published by the Free Software Foundation. This file
is distributed without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose.

See "COPYING" in the source distribution for more information.

Headers in this file shall remain intact.
'''

from gi.repository import GObject
import pyatspi
import sys

class SequenceStep(GObject.GObject):
  '''
  Base class for all sequence steps in a Macaroon sequence. Emits a "done"
  signal when the step is done.

  @cvar delta_time: Time, in milliseconds, before this step should be executed.
  @type delta_time: integer

  @ivar done: True if step is done.
  @type done: boolean
  '''
  __gsignals__ = {'done' : (GObject.SignalFlags.RUN_FIRST,
                                 None, ())}
  delta_time = 0
  def __init__(self):
    '''
    Initialize L{SequenceStep}.
    '''
    super(SequenceStep, self).__init__()
    self.done = False

  def stepDone(self):
    '''
    Puts instance in "done" state. And emits "done" signal.

    @return: Return False because this is usually called through a
    gobject timeout.
    @rtype: boolean
    '''
    if not self.done:
      self.done = True
      self.emit('done')
    return False

class AtomicAction(SequenceStep):
  '''
  A L{SequenceStep} that performs a certain action.

  @ivar _func: Function to call.
  @type _func: callable
  @ivar _args: Arguments to give the function.
  @type _args: list
  @ivar _kwargs: Key word arguments for function.
  @type _kwargs: dictionary
  '''
  def __init__(self, delta_time, func, *args, **kwargs):
    '''
    Initialize L{AtomicAction}.

    @param delta_time: Time, in milliseconds, before this step should be
    executed.
    @type delta_time: integer
    @param func: Function to call.
    @type func: callable
    @param _args: Arguments to give the function.
    @type _args: list
    @param _kwargs: Key word arguments for function.
    @type _kwargs: dictionary
    '''
    SequenceStep.__init__(self)
    self.delta_time = delta_time
    self._func = func
    self._args = args
    self._kwargs = kwargs

  def __call__(self):
    '''
    Perform the given function.
    '''
    self._func(*self._args, **self._kwargs)
    self.stepDone()

class DebugAction(AtomicAction):
  '''
  An action that prints a debug message to standard output.
  '''
  def __init__(self, debug_msg, delta_time=0):
    '''
    Initialize L{DebugAction}

    @param debug_msg: Message to print out to standard output.
    @type debug_msg: string
    @param delta_time: Time to wait before printing the message.
    @type delta_time: integer
    '''
    self._debug_msg = debug_msg
    AtomicAction.__init__(self, 0, self._printDebugMsg)

  def _printDebugMsg(self):
    '''
    Print the debug message.
    '''
    print(self._debug_msg)

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return 'Debug message: %s' % self._debug_msg

class PauseAction(AtomicAction):
  '''
  A simple pause for a given amount of time.
  '''
  def __init__(self, delta_time):
    '''
    Initialize a L{PauseAction}.

    @param delta_time: Duration of pause in milliseconds.
    @type delta_time: integer
    '''
    AtomicAction.__init__(self, delta_time, lambda args, kwargs: None, [], {})

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return 'Pause for %d milliseconds' % self.delta_time


class CallableAction(AtomicAction):
  '''
  A simplified L{AtomicAction} that makes it easy for simply calling a function.
  '''
  def __init__(self, func, *args, **kwargs):
    '''
    Initialize L{AtomicAction}.

    @param delta_time: Time, in milliseconds, before this step should be
    executed.
    @type delta_time: integer
    @param func: Function to call.
    @type func: callable
    @param _args: Arguments to give the function.
    @type _args: list
    @param _kwargs: Key word arguments for function.
    @type _kwargs: dictionary
    '''
    AtomicAction.__init__(self, 0, func, *args, **kwargs)



'''
Actions used for waiting for something or for a certain time.

@author: Eitan Isaacson
@copyright: Copyright (c) 2007 Eitan Isaacson
@license: GPL

This file may be distributed and/or modified under the terms of the GNU General
Public License version 2 as published by the Free Software Foundation. This file
is distributed without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose.

See "COPYING" in the source distribution for more information.

Headers in this file shall remain intact.
'''

from gi.repository import GLib
import pyatspi

import re

from .sequence_step import SequenceStep
from . import utils
import sys

_ = lambda x: x

class WaitAction(SequenceStep):
  '''
  Base class for all wait actions.

  @ivar wait_for: List of events to wait for.
  @type wait_for: list of string
  @ivar _acc_name: Name of source accessible of events we are waiting for.
  @type _acc_name: string
  @ivar _acc_path: Tree path in application to source accessible.
  @type _acc_path: list of integer
  @ivar _acc_role: Role of source accessible.
  @type _acc_role: integer
  @ivar _timeout: Time to wait in milliseconds before timing out.
  @type _timeout: integer
  '''
  def __init__(self, event, acc_name, acc_path, acc_role, timeout):
    '''
    Initialize a L{WaitAction}

    @param event: Event to wait for
    @type event: string
    @param acc_name: Name of source accessible of events we are waiting for.
    @type acc_name: string
    @param acc_path: Tree path in application to source accessible.
    @type acc_path: list of integer
    @param acc_role: Role of source accessible.
    @type acc_role: integer
    @param timeout: Time to wait in milliseconds before timing out.
    @type timeout: integer
    '''
    SequenceStep.__init__(self)
    if isinstance(event, list):
      self.wait_for = event
    else:
      self.wait_for = [event]
    self._acc_name = acc_name
    self._acc_path = acc_path
    self._acc_role = acc_role
    self._timeout = timeout

  def __call__(self, cached_events):
    '''
    Called when an insance is "called".

    @param cached_events: List of events that have been cached since the
    last sequence step.
    @type cached_events: list of Accessibility.Event
    '''
    if self.wait_for == []: return
    self._cached_events = cached_events
    if self.checkExistingState(): return
    pyatspi.Registry.registerEventListener(self.onEvent, *self.wait_for)
    GLib.timeout_add(self._timeout, self._onTimeout)

  def checkExistingState(self):
    '''
    Check to see if perhaps the event we are waiting for has already been fired
    before this instance has been created.

    @return: True if the event has already occured.
    @rtype: boolean
    '''
    if not self.done:
      for event in self._cached_events:
        self.onEvent(event)
    return self.done

  def _onTimeout(self):
    '''
    Timout callback.
    '''
    if not self.done:
      sys.stderr.write('Macaroon timeout: %s\n' % self)
    self.stepDone()
    return False

  def onEvent(self, event):
    '''
    Event callback.

    @param event: Event that has triggered this callback.
    @type event: Accessibility.Event
    '''
    if (self._acc_name is None or \
          self._acc_name == event.source.name) and \
       (self._acc_path is None or \
          self._acc_path == pyatspi.getPath(event.source)) and \
       (self._acc_role is None or \
          self._acc_role == event.source.getRole()):
      self.stepDone()

  def stepDone(self):
    '''
    Done with this sequence step. Deregister event listeners.
    '''
    if not self.done:
      pyatspi.Registry.deregisterEventListener(self.onEvent, *self.wait_for)
    return SequenceStep.stepDone(self)

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return 'Wait for %s' % ','.join(self.wait_for)

class WaitForWindowActivate(WaitAction):
  '''
  Wait for a window to become activated.

  @todo: Eitan: Provide the ability to match regular expressions since frame
  titles are highly dynamic.

  @ivar _frame_re: Window name to match.
  @type _frame_re: string
  @ivar _application_re: Application name to match, or None if no
  matching needs to be performed.
  @type _application_re: string
  '''
  def __init__(self, frame_re, application_re=None, timeout=30000):
    '''
    Initialize L{WaitForWindowActivate}.

    @param frame_re: Window name to match.
    @type frame_re: string
    @param application_re: Application name to match, or None if no
    matching needs to be performed.
    @type application_re: string
    @param timeout: Time to wait in milliseconds before timing out.
    @type timeout: integer
    '''
    WaitAction.__init__(self,
                        ['window:activate',
                         'object:property-change:accessible-name'],
                        None, None, None, timeout)
    self._frame_re = frame_re
    self._application_re = application_re

  def checkExistingState(self):
    '''
    Check if the window we are waiting for is already activated.

    @return: True if window is already activated.
    @rtype: boolean
    '''
    self._active_frame = utils.getActiveFrame()
    if self.isRightFrame(self._active_frame):
      self.stepDone()
    return WaitAction.checkExistingState(self)

  def onEvent(self, event):
    '''
    Event callback. Check if this is the fame we are looking for and
    move to the next step.
    '''
    if event.type == 'window:activate' or \
          (event.type == 'object:property-change:accessible-name' and \
             event.source == self._active_frame):
      self._active_frame = event.source
      if self.isRightFrame(self._active_frame):
        self.stepDone()

  def isRightFrame(self, acc):
    '''
    Check if this is the frame we are waiting for.

    @param acc: Frame accessible.
    @type acc: Accessibility.Accessible

    @return: True if this is the right frame.
    @rtype: boolean
    '''
    return acc is not None and \
      (acc.name == self._frame_re or \
      re.compile(self._frame_re and self._frame_re or "").match(acc.name))

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return _('Wait for window %s to get activated') % self._frame_re

class WaitForFocus(WaitAction):
  '''
  Wait for a focus event.
  '''
  def __init__(self, acc_name=None,
               acc_path=None, acc_role=None, timeout=10000):
    '''
    Initialize a L{WaitForFocus}

    @param acc_name: Name of source accessible of events we are waiting for.
    @type acc_name: string
    @param acc_path: Tree path in application to source accessible.
    @type acc_path: list of integer
    @param acc_role: Role of source accessible.
    @type acc_role: integer
    @param timeout: Time to wait in milliseconds before timing out.
    @type timeout: integer
    '''
    WaitAction.__init__(self, "focus:", acc_name, acc_path, acc_role, timeout)

  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    if self._acc_name and self._acc_role:
      identifier = '%s called "%s"' % \
          (repr(self._acc_role).replace('ROLE_','').lower().replace('_',' '),
           self._acc_name)
    elif self._acc_name:
      identifier = self._acc_name
    elif self._acc_role:
      identifier = \
          repr(self._acc_role).replace('ROLE_','').lower().replace('_',' ')
    elif self._acc_path:
      identifier = 'componenet at %s' % self._acc_path
    else:
      identifier = 'anything'
    return _('Wait for %s to be focused') % identifier

class WaitForDocLoad(WaitAction):
  '''
  Wait for a document to load.
  '''
  def __init__(self):
    '''
    Initialize L{WaitForDocLoad}
    '''
    WaitAction.__init__(self, 'document:load-complete',
                        None, None, None, 30000)
  def __str__(self):
    '''
    String representation of instance.

    @return: String representation of instance.
    @rtype: string
    '''
    return 'Wait for document to load'



'''
Defines behavior of the accessible tree view widget.

@author: Peter Parente
@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''

from gi.repository import Gio as gio
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import GObject

import pyatspi
import os
from . import menus
from time import sleep
from .icons import getIcon
from .node import Node
from .tools import ToolsAccessor, getTreePathBoundingBox
from .i18n import _

COL_ICON = 0
COL_NAME = 1
COL_ROLE = 2
COL_CHILDCOUNT = 3
COL_FILLED = 4
COL_DUMMY = 5
COL_ACC = 6

class AccessibleModel(gtk.TreeStore, ToolsAccessor):
  '''
  Stores the desktop accessible tree. Only populates sections of the tree
  that are being viewed. This cuts short on a lot of potential overhead.

  @ivar desktop: The desktop accessible. It holds references to all the
  application L{Accessibility.Accessible}s
  @type desktop: L{Accessibility.Accessible}
  @ivar acc_cache: A list of L{Accessibility.Accessible}s that are currently
  resident in the model. This helps with faster searching.
  @type acc_cache: list
  '''
  __gsignals__ = {'row-filled' :
                  (GObject.SignalFlags.RUN_FIRST,
                   None,
                   (GObject.TYPE_PYOBJECT,)),
                  'start-populating' :
                    (GObject.SignalFlags.RUN_FIRST,
                     None,
                     ()),
                  'end-populating' :
                    (GObject.SignalFlags.RUN_FIRST,
                     None,
                     ())}

  def __init__(self, desktop_acc):
    '''
    Initializes the L{AccessibleModel} and all of the needed instant variables.
    Connects required signals.
    '''
    self.acc_cache = [desktop_acc]
    gtk.TreeStore.__init__(self, GdkPixbuf.Pixbuf, str, str, int, bool, bool, object)
    self.connect('row-changed', self._onRowChanged)
    self.connect('row-filled', self._onRowFilled)
    self.desktop = desktop_acc
    self._path_to_populate = None
    self._populating_tasks = 0
    self.isPopulating = False

  def _onRowChanged(self, model, path, iter):
    '''
    A callback on "row-changed" that pre-populates a given row when it changes.

    @param model: This model.
    @type model: L{AccessibleModel}
    @param path: The path to the row that changed.
    @type path: tuple
    @param iter: the iter of the row that changed.
    @type iter: L{gtk.TreeIter}
    '''
    self._prepopLevel(model[iter][COL_ACC], iter)

  def insert(self, iter, index, row):
    '''
    A method that overrides the L{gtk.TreeStore} insert method.
    Pre-populate and add the new accessible to acc_cache.

    @param iter: The parent iter of the newly inserted row.
    @type iter: L{gtk.TreeIter}
    @param index: The place in which to insert the new row.
    @type index: integer
    @param row: A list of columns o insert into the row,
    @type row: list

    @return: The newly created iter of the inserted row.
    @rtype: L{gtk.TreeIter}
    '''
    acc = row[COL_ACC]
    if acc:
      self.acc_cache.append(acc)
    new_iter = gtk.TreeStore.insert(self, iter, index, row)
    self._prepopLevel(acc, new_iter)

  def append(self, iter, row):
    '''
    A method that overrides the L{gtk.TreeStore} append method.
    Pre-populate and add the new accessible to acc_cache.

    @param iter: The parent iter of the newly inseted row.
    @type iter: L{gtk.TreeIter}
    @param row: A list of columns o insert into the row,
    @type row: list

    @return: The newly created iter of the inserted row.
    @rtype: L{gtk.TreeIter}
    '''
    self.insert(iter, -1, row)

  def remove(self, iter):
    '''
    A method that overrides the L{gtk.TreeStore} remove method.
    Remove the row's accessible from acc_cache.

    @param iter: The parent iter of the newly inserted row.
    @type iter: L{gtk.TreeIter}

    @return: True if L{iter} is still valid.
    @rtype: boolean
    '''
    if self[iter][COL_ACC]:
      self.acc_cache.remove(self[iter][COL_ACC])
    return gtk.TreeStore.remove(self, iter)

  def isInModel(self, acc):
    '''
    Checks if the given L{Accessibility.Accessible} is resident in the model.

    @param acc: The L{Accessibility.Accessible} to check.
    @type acc: L{Accessibility.Accessible}

    @return: True if it is in the model.
    @rtype: boolean
    '''
    return acc in self.acc_cache

  def getChildrenAccs(self, iter):
    '''
    Get list of accessible children of a given row.

    I{Note:} This method returns the accessible's children as they currently exist in
    the model. The list of accessibles is not necessarily identical to the actual
    children of the accessible at the given row.

    @param iter: Th iter of the row that we want it's children
    @type iter: L{gtk.TreeIter}

    @return: List of children
    @rtype: list
    '''
    if iter:
      return [row[COL_ACC] for row in self[iter].iterchildren()]
    else:
      return [row[COL_ACC] for row in self]

  def _prepopLevel(self, parent, iter):
    '''
    Pre-populate a row. If a L{Accessibility.Accessible} of the given row has children,
    we need to add to it one dummy child row so that the expander will show and
    enable the user to expand it. We populate the children rows at expansion time.

    @param parent: The given row's accessible.
    @type parent: L{Accessibility.Accessible}
    @param iter: Th iter of the row that needs to be pre-populated
    @type iter: L{gtk.TreeIter}
    '''
    if (parent and self.children_number(parent) > 0 and
        not self.isMyApp(parent) and self.iter_n_children(iter) == 0):
      row = self._buildRow(None, True)
      self.append(iter, row)

  def popLevel(self, iter):
    '''
    Populate a row with children rows, according to the row's accessible's children.

    @param iter: Th iter of the row that needs to be populated
    @type iter: L{gtk.TreeIter}
    '''
    if iter:
      tree_path = self.get_path(iter)
      path = tuple(tree_path.get_indices())

      row_reference = gtk.TreeRowReference.new(self, tree_path)
      #row_reference = gtk.TreeRowReference()
    else:
      row_reference = None
    self._populating_tasks += 1
    if self._populating_tasks == 1:
      self.emit('start-populating')
    GLib.idle_add(self._popOnIdle, row_reference)

  def _childrenIndexesInParent(self, accessible):
    '''
    Gets ids from children (either all of them or only those that are not leaves).

    @param accessible: An accessible from which we want to get children ids.
    @type accessible: L{Accessibility.Accessible}

    @return: A list with children ids.
    @rtype: list
    '''
    return [i for i in range(accessible.childCount)]

  def _selectChild(self, parent, index):
    '''
    Returns a child id.

    @param parent: An accessible from which we want to select a child id.
    @type parent: L{Accessibility.Accessible}
    @param index: An index to retrieve the child id. In some cases, it is equal to the id.
    @type index: integer

    @return: A child id.
    @rtype: integer
    '''
    return index

  def children_number(self, parent):
    '''
    Returns how many children an accessible has (either all of them or only those that are not leaves).

    @param parent: An accessible from which we want to get the number of children.
    @type parent: L{Accessibility.Accessible}

    @return: The number of children (including leaves or not) an accessible has.
    @rtype: integer
    '''
    return parent.childCount

  def _popOnIdle(self, row_reference):
    '''
    Idle callback for populating the children of a given row reference.

    @param row_reference: The parent row reference.
    @type row_reference: gtk.TreeRowReference

    @return: False if task is done (to stop handler).
    @rtype: boolean
    '''
    remove_iter = None
    iter = None
    parent = self.desktop

    if row_reference:
      if not row_reference.valid():
        self._endPopTask()
        return False
      iter = self.get_iter(row_reference.get_path())
      parent = self[iter][COL_ACC]
      tree_path = row_reference.get_path()
      path = tuple(tree_path.get_indices())
      if self[path+(0,)][COL_DUMMY]:
        remove_iter = self.iter_children(iter)

    already_populated_num = self.iter_n_children(iter)

    if already_populated_num >= self.children_number(parent) and \
          not remove_iter:
      if iter:
        self[iter][COL_FILLED] = True
      self.emit('row-filled', iter)
      self._endPopTask()
      return False
    elif remove_iter:
      already_populated_num -= 1
    try:
      index = self._selectChild(parent, already_populated_num)
      child = parent.getChildAtIndex(index)
    except LookupError:
      child = None

    row = self._buildRow(child)
    self.append(iter, row)

    if remove_iter:
      self.remove(remove_iter)
    return True

  def _endPopTask(self):
    '''
    Convinience function for stopping a populating task.
    '''
    self._populating_tasks -= 1
    if self._populating_tasks == 0:
      self.emit('end-populating')

  def popToPath(self, path):
    '''
    Populate the model with accessible nodes up to given path.

    @param path: Path to populate model to.
    @type path: tuple
    '''
    if not self._walkThroughFilled(path):
      self._path_to_populate = path
    else:
      self.emit('row-filled', self.get_iter(path))
      self._path_to_populate = None

  def _walkThroughFilled(self, path):
    '''
    Reach node in path that is not populated yet, and populate it.

    @param path: Path to fill.
    @type path: tuple

    @return: True if entire path is populated.
    @rtype: boolean
    '''
    for i in range(1, len(path)):
      if not self[path[:i]][COL_FILLED]:
        self.popLevel(self.get_iter(path[:i]))
        return False
    return True

  def _onRowFilled(self, model, iter):
    '''
    Callback for "row-filled" signal. If there is a specific path we need to populate,
    we continue populating the next node.

    @param model: Model that emitted signal (self).
    @type model: L{AccessibleModel}
    @param iter: Iter of row that has been populated.
    @type iter: gtk.TreeIter
    '''
    if iter and self._path_to_populate:
      tree_path = self.get_path(iter)
      path = tuple(tree_path.get_indices())
      if self._path_to_populate[:len(path)] == path:
        if self._walkThroughFilled(self._path_to_populate):
          self._path_to_populate = None

  def getIndexInParent(self, child):
    '''
    Returns a position of a child in its parent.

    @param child: Accessible for which we want to determine the index.
    @type child: L{Accessibility.Accessible}

    @return: The child's index or -1 if it wasn't found
    @rtype: integer
    '''
    return child.getIndexInParent()

  def getAccPath(self, acc):
    '''
    Get the tree path that a given accessible should have.

    I{Note:} The accessible does not necessarily need to be resident in the model.

    @param acc: The accessible we want a path of.
    @type acc: L{Accessibility.Accessible}

    @return: The path to the accessible.
    @rtype: tuple
    '''
    path = ()
    child = acc
    while child.get_parent():
      try:
        index_in_parent = self.getIndexInParent(child)
        if index_in_parent < 0:
          break
        path = (index_in_parent,) + path
      except Exception as e:
        return None
      child = child.get_parent()

    try:
      path = (list(self.desktop).index(child),) + path
    except Exception as e:
      return None

    return path

  def _buildRow(self, accessible, dummy=False):
    '''
    Wrapper for building a row in the tree. Use this method instead of trying
    to construct the row by hand as it will be synced with the design of the
    model fields.

    @param accessible: Accessible object
    @type accessible: L{Accessibility.Accessible}
    @param dummy: Is this a dummy row?
    @type dummy: boolean
    '''
    if accessible is not None:
      icon = getIcon(accessible)
      if self.isMyApp(accessible):
        name = "Accerciser"
        role = "accerciser"
        count = 0
      else:
        name = accessible.name
        role = accessible.getLocalizedRoleName()
        count = accessible.childCount
    else:
      icon = None
      if not dummy:
        name = _('<dead>')
      else:
        name = None
      role = None
      count = 0
    return [icon, name, role, count, False, dummy, accessible]

class AccessibleTreeView(gtk.TreeView, ToolsAccessor):
  '''
  The treeview for the desktop's accessibles. The treeview's model (L{AccessibleModel}
  is only populated when the treeview is traversed and nodes are expanded. This class
  listens for 'row-expanded' events in order to have a tree node populated. Nodes that
  are selected are updated into the L{Node} instance variable. This treeview also
  updates automatically in response to at-spi children change events.

  @ivar desktop: The desktop accessible. It holds references to all the
  application L{Accessibility.Accessible}s
  @type desktop: L{Accessibility.Accessible}
  @ivar application: Main application.
  @type application: gtk.Application
  @ivar node: An object with a reference to the currently selected accessible.
  @type node: L{Node}
  @ivar model: The data model of this treeview.
  @type model: L{AccessibleModel}
  '''
  def __init__(self, application, node):
    '''
    Initialize the treeview. Build the proper columns.
    Connect all of the proper signal handlers and at-spi event handlers.

    @param application: Main application.
    @type application: gtk.Application
    @param node: The main application node.
    @type node: L{Node}
    '''
    gtk.TreeView.__init__(self)

    self.application = application
    self.desktop = pyatspi.Registry.getDesktop(0)
    self.node = node
    self.node.update(self.desktop)
    self._changed_handler = self.node.connect('accessible_changed',
                                             self._onAccChanged)
    self.connect('row-activated', self._onRowActivated)

    self.model = AccessibleModel(self.desktop)
    self.filter = self.model.filter_new()
    self.filter.set_visible_func(self._filterNoChildApps, data=None)
    self.set_model(self.filter)

    crt = gtk.CellRendererText()
    crp = gtk.CellRendererPixbuf()
    tvc = gtk.TreeViewColumn(_('Name'))
    tvc.pack_start(crp, False)
    tvc.pack_start(crt, True)
    tvc.add_attribute(crp, 'pixbuf', COL_ICON)
    tvc.add_attribute(crt, 'text', COL_NAME)
    tvc.set_resizable(True)
    self.append_column(tvc)

    crt= gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Role'))
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', COL_ROLE)
    tvc.set_resizable(True)
    self.append_column(tvc)

    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn(_('Children'))
    tvc.pack_start(crt, True)
    tvc.add_attribute(crt, 'text', COL_CHILDCOUNT)
    tvc.set_resizable(True)
    self.append_column(tvc)

    self.model.connect('row-filled', self._onRowFilled)
    self.model.connect('start-populating', self._onStartPop)
    self.model.connect('end-populating', self._onEndPop)
    self._path_to_expand = None

    self._refreshTopLevel()

    selection = self.get_selection()
    selection.unselect_all()
    selection.connect('changed', self._onSelectionChanged)
    self.connect('row-expanded', self._onExpanded)

    pyatspi.Registry.registerEventListener(self._accEventChildChanged,
                                           'object:children-changed')

    pyatspi.Registry.registerEventListener(
        self._accEventNameChanged,
        'object:property-change:accessible-name')

    self._hide_leaves = True

    menu, name, label, callback = (menus.view_menu_treeview_section,
                                   'hide_show_leaves', _('_Show Applications without children'),
                                   self._onHideShowLeaves)
    action_name = 'app.' + name
    menu_item = gio.MenuItem.new(label, action_name)
    menu.append_item(menu_item)
    action = gio.SimpleAction.new_stateful(name, None, GLib.Variant.new_boolean(False))
    action.connect('change-state', callback)
    self.application.add_action(action)

    self.application.addMenuItem(menus.view_menu_treeview_section, 'refresh_all', 'view-refresh', _('_Refresh Registry'),
        None, self._refreshTopLevel)

    # add "Refresh Node" to both, the menu bar and the tree view context menu
    menu_item_refresh_current, self.refresh_current_action = \
      self.application.addMenuItem(menus.view_menu_treeview_section,
                                  'refresh_current', 'go-jump', _('Refresh _Node'),
                                  None, self._refreshCurrentLevel)
    self.refresh_current_action.set_property('enabled', False)
    menus.treeview_context_menu.append_item(menu_item_refresh_current)

    self.connect('popup-menu', self._onPopup)
    self.connect('button-press-event', self._onPopup)
    self.connect('key-press-event', self._onKeyPress)

    self.connect('cursor-changed', self._onCursorChanged)

  def _filterNoChildApps(self, model, iter, data):
    '''
    Filter all rows of applications without children
    '''
    return not self._hide_leaves or model.iter_parent(iter) != None \
      or model[iter][COL_CHILDCOUNT] != 0 or self.isMyApp(model[iter][COL_ACC])

  def _onCursorChanged(self, tree):
    '''
    Set sensitivity of refresh function only if the tree cursor is
    on an accessible.
    '''
    path = self.get_cursor()[0]
    self.refresh_current_action.set_property('enabled', path is not None)

  def _onKeyPress(self, w, event):
    '''
    Expand or collapse a row on Left/Right key-press
    '''
    if event.state & \
        (gdk.ModifierType.SHIFT_MASK
        |gdk.ModifierType.CONTROL_MASK
        |gdk.ModifierType.MOD1_MASK) == 0:
      path, col = self.get_cursor()
      if path is not None:
        if event.keyval == gdk.KEY_Left:
          if not self.collapse_row(path):
            # if we ccouldn't collapse the current row, go to the parent
            if path.up():
              self.set_cursor(path)
          return True
        elif event.keyval == gdk.KEY_Right:
          self.expand_row(path, False)
          return True
    return False

  def _onPopup(self, w, event=None):
    '''
    Callback for popup button or right mouse button. Brings up a context
    menu.
    '''
    if event:
      if event.button != 3:
        return False
      info = self.get_path_at_pos(int(event.x), int(event.y))
      if info is None:
        return False
      path = info[0]
      if self.isMyApp(self.filter[self.filter.get_iter(path)][COL_ACC]):
        return False
      selection = self.get_selection()
      selection.set_mode(gtk.SelectionMode.NONE)
      self.set_cursor(path, None, False)
      selection.set_mode(gtk.SelectionMode.SINGLE)
      time = event.time
      button = event.button
      func = None
      extra_data = None
    else:
      path, col= self.get_cursor()
      time = gtk.get_current_event_time()
      button = 0
      extra_data = getTreePathBoundingBox(self, path, col)
      func = lambda m, b: (b.x, b.y + (b.height/2), True)

    menu = gtk.Menu.new_from_model(menus.treeview_context_menu)
    # attach to widget, since actions are searched in the widget's hierarchy,
    # which includes the the application window where the actions are defined
    # s. https://docs.gtk.org/gtk3/ctor.Menu.new_from_model.html
    menu.attach_to_widget(self, None)
    menu.popup(None, None, func, extra_data, button, time)
    return True

  def _refreshTopLevel(self, action=None, data=None):
    '''
    Refreshes the entire tree at the desktop level.

    @param action: Action object that emitted this signal, if any.
    @type: gio.Action
    '''
    self.model.clear()
    self.model.popLevel(None)
    # iter over all apps in the desktop too

  def _refreshCurrentLevel(self, action, data=None):
    '''
    Refreshes the current level. Selects and expands the parent of the level.

    @param action: Action object that emitted this signal, if any.
    @type: gio.SimpleAction
    '''
    path = self.get_cursor()[0]
    if path == None:
      return
    filter_iter = self.filter.get_iter(path)
    if self.isMyApp(self.filter[filter_iter][COL_ACC]):
      return
    is_expanded = self.row_expanded(path)
    self._refreshChildren(self.filter.convert_iter_to_child_iter(filter_iter))
    if is_expanded:
      self.expand_row(path, False)
      self._onExpanded(self, self.filter.get_iter(path), path)

  def _onHideShowLeaves(self, action, value, data=None):
    '''
    Callback for toggling showing applications without children.

    @param action: Action object that emitted callback.
    @type action: gio.SimpleAction
    '''
    action.set_state(value)
    self._hide_leaves = not value.get_boolean()
    self.filter.refilter()

  def _onExpanded(self, treeview, iter, path):
    '''
    Populates a level when it is expanded. Removes the previously added dummy
    node.

    @param treeview: The L{AccessibleTreeView} that emitted the signal.
    @type treeview: L{AccessibleTreeView}
    @param iter: The iter that has been expanded.
    @type iter: L{gtk.TreeIter}
    @param path: The path to the row that has been expanded.
    @type path: tuple
    '''
    # don't repopulate if it has been filled before
    if self.filter[iter][COL_FILLED]:
      return
    # populate this level
    self.model.popLevel(self.filter.convert_iter_to_child_iter(iter))

  def _accEventNameChanged(self, event):
    '''
    Event handler for "object:property-change:accessible-name".
    Updates the treeview accordingly.

    @param event: The event which triggered this handler.
    @type event: L{pyatspi.event.Event}
    '''
    if self.isMyApp(event.source) or event.source == self.desktop:
      # Bad karma
      return
    if self.model.isInModel(event.source):
      try:
        path = self.model.getAccPath(event.source)
        iter = self.model.get_iter(path)
      except:
          pass
      else:
          self.model[iter][COL_NAME] = event.source.name

  def _accEventChildChanged(self, event):
    '''
    Event handler for "object:children-changed". Updates the treeview accordingly.

    @param event: The event which triggered this handler.
    @type event: L{pyatspi.event.Event}
    '''
    if self.isMyApp(event.source):
      # Bad karma
      return
    if self.model.isInModel(event.source):
      try:
        path = self.model.getAccPath(event.source)
        iter = self.model.get_iter(path)
      except:
        iter = None
        path = None
      if (event.source == self.desktop) or \
            (path and self.model[path][COL_FILLED]):
        if event.type.minor == 'add':
          self._addChild(iter, event.source)
        elif event.type.minor == 'remove':
          self._removeChild(iter, event.source)
      if iter and self.model.iter_is_valid(iter):
        self.model[iter][COL_CHILDCOUNT] = event.source.childCount

  def _addChild(self, iter, parent):
    '''
    Add the new child to the given accessible.

    @param iter: Th iter of the row that needs a child added.
    @type iter: L{gtk.TreeIter}
    @param parent: The given row's accessible.
    @type parent: L{Accessibility.Accessible}
    '''
    old_children = set(self.model.getChildrenAccs(iter))
    new_children = set(list(parent))

    added = new_children.difference(old_children)
    try:
      new_child = added.pop()
    except KeyError:
      return
    row = self.model._buildRow(new_child)
    if new_child is None:
      self.model.append(iter, row)
    else:
      self.model.insert(iter, self.model.getIndexInParent(new_child), row)
      # We do this because an application won't have an icon loaded in
      # the window manager when it is first registered to at-spi
      if new_child == new_child.getApplication():
        GLib.timeout_add(1000, self._refreshIcon, new_child)

  def _refreshIcon(self, app):
    '''
    Refresh the icon of a given application's accessible. This is done because it
    takes wnck a while to load an application's icon at application startup.

    @param app: The given application's accessible.
    @type app: L{Accessibility.Accessible}
    '''
    path = self.model.getAccPath(app)
    try:
      self.model[path][COL_ICON] = getIcon(app)
    except:
      pass
    return False

  def _removeChild(self, parent_iter, parent):
    '''
    Remove a child from the given accessible node.

    @param parent_iter: Th iter of the row that needs a child removed.
    @type parent_iter: L{gtk.TreeIter}
    @param parent: The given row's accessible.
    @type parent: L{Accessibility.Accessible}
    '''
    if parent_iter:
      iter = self.model.iter_children(parent_iter)
    else:
      iter = self.model.get_iter_first()

    while iter:
      if self.model[iter][COL_ACC] not in parent:
        cursor_path = self.get_cursor()[0]
        if cursor_path != None:
          (res, filter_iter) = self.filter.convert_child_iter_to_iter(iter)
          if res:
            filter_path = self.filter.get_path(filter_iter)
            if filter_path.is_ancestor(cursor_path):
              cursor_path = filter_path
            if 0 == filter_path.compare(cursor_path):
              if filter_path.prev() or filter_path.up():
                self.set_cursor(filter_path, None, False)
        if not self.model.remove(iter):
          break
      else:
        iter = self.model.iter_next(iter)

  def _refreshChildren(self, iter):
    '''
    Remove all of a given node's children from the model.

    @param iter: The parent node.
    @type iter: L{gtk.TreeIter}
    '''
    if not iter:
      self._refreshTopLevel()
      return
    child_iter = self.model.iter_children(iter)
    while child_iter:
      if not self.model.remove(child_iter):
        break
    acc = self.model[iter][COL_ACC]
    self.model[iter][COL_CHILDCOUNT] = acc.childCount
    self.model[iter][COL_FILLED] = False

  def refreshSelected(self):
    '''
    Manually refresh the selected node.
    '''
    selection = self.get_selection()
    model, iter = selection.get_selected()
    if not self.isMyApp(model[iter][COL_ACC]):
      self._refreshChildren(iter)

  def _onSelectionChanged(self, selection):
    '''
    Update the accessible according to the selected row.

    @param selection: The selection object that emitted the the 'selection-changed'
    signal.
    @type selection: L{gtk.TreeSelection}
    '''
    self._path_to_expand = None
    model, iter = selection.get_selected()
    if iter:
      new_acc = model[iter][COL_ACC]
    else:
      new_acc = self.desktop
    if new_acc == self.node.acc:
      return
    self.node.handler_block(self._changed_handler)
    self.node.update(new_acc)
    self.node.handler_unblock(self._changed_handler)
    if iter:
      tree_path = model.get_path(iter)
      path = tuple(tree_path.get_indices())

      self.node.tree_path = list(path[1:])

  def _onAccChanged(self, node, acc):
    '''
    Change the treeview's selection to the updated accessible in the L{Node}.

    @param node: The L{node} that emitted the signal.
    @type node: L{Node}
    @param acc: The new accessible in the node.
    @type acc: L{Accessibility.Accessible}
    '''
    if self.isMyApp(acc):
      # Bad karma
      return
    path = self.model.getAccPath(acc)
    if not path:
      return
    if len(path) >= 1:
      self.selectNodeAtPath(path)
      self.node.tree_path = list(path[1:])

  def selectNodeAtPath(self, path):
    '''
    Select the node at the current path. The path does not need to exist in the model,
    only in the accessibles tree. The model will get populated accordingly.

    @param path: The path to select.
    @type path: tuple
    '''
    try:
      dummy = self.model[path][COL_DUMMY]
    except:
      dummy = True
    if dummy:
      self._path_to_expand = path
      self.model.popToPath(path)
      # process pending events until the path actually
      # got populated via the corresponding idles
      main_loop_context = GLib.MainContext.default()
      while self.isPopulating and main_loop_context.iteration(False):
        pass
      try:
        dummy = self.model[path][COL_DUMMY]
      except:
        pass
    if not dummy:
      self._selectExistingPath(path)

  def _onRowFilled(self, model, iter):
    '''
    Callback for "row-filled" (populated) signal. Used for selecting a child node in
    the given iter if L{selectNodeAtPath} was called on one of the given
    iter's children.

    @param model: Model that emitted this signal.
    @type model: L{AccessibleModel}
    @param iter: Iter of row that was populated with children.
    @type iter: gtk.TreeIter
    '''
    if iter and self._path_to_expand and \
          self._path_to_expand[:-1] == model.get_path(iter).get_indices():
      self._selectExistingPath(self._path_to_expand)
      self._path_to_expand = None

  def _selectExistingPath(self, model_path):
    '''
    Select a path that already exists. Expand, scroll, and select.

    @param model_path: Path to select, referring to the unfiltered model.
    @type path: tuple
    '''
    iter = self.model.get_iter(model_path)
    (res, filter_iter) = self.filter.convert_child_iter_to_iter(iter)
    if not res:
        return
    filter_path = self.filter.get_path(filter_iter)
    tree_path = gtk.TreePath(filter_path[:-1])
    if len(filter_path) > 1:
      self.expand_to_path(tree_path)
    self.scroll_to_cell(filter_path)
    selection = self.get_selection()
    selection.select_path(filter_path)


  def _onStartPop(self, model):
    '''
    Callback for when the model is populating, changes the cursor to a watch.

    @param model: Model that emitted the signal.
    @type model: L{AccessibleModel}
    '''
    self.isPopulating = True
    if self.get_window():
      window = self.get_window()
      window.set_cursor(gdk.Cursor(gdk.CursorType.WATCH))

  def _onEndPop(self, model):
    '''
    Callback for when the model stops populating, changes the cursor to an arrow.

    @param model: Model that emitted the signal.
    @type model: L{AccessibleModel}
    '''
    if self.get_window():
      window = self.get_window()
      window.set_cursor(gdk.Cursor(gdk.CursorType.TOP_LEFT_ARROW))
    self.isPopulating = False

  def _onRowActivated(self, treeview, path, view_column):
    '''
    When the row is activated (double clicked, or enter), blink the selected
    accessible, if possible.

    @param treeview: The L{AccessibleTreeView} that emitted the signal.
    @type treeview: L{AccessibleTreeView}
    @param path: The path to the selected row.
    @type path: tuple
    @param view_column: The column in the activated row.
    @type view_column: L{gtk.TreeViewColumn}
    '''
    if not self.isMyApp(self.filter[self.filter.get_iter(path)][COL_ACC]):
      self.node.highlight()



# Macaroon - a desktop macro tool
# Copyright (C) 2007 Eitan Isaacson <eitan@ascender.com>
# All rights reserved.

# This file may be distributed and/or modified under the terms of
# the GNU General Public License version 2 as published by
# the Free Software Foundation.
# This file is distributed without any warranty; without even the implied
# warranty of merchantability or fitness for a particular purpose.
# See "COPYING" in the source distribution for more information.

# Headers in this file shall remain intact.



'''
AT-SPI validation plugin.

@author: Peter Parente
@organization: IBM Corporation
@copyright: Copyright (c) 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
import gi

from gi.repository import Gtk as gtk
from gi.repository import GObject
from gi.repository import GLib

import os
import traceback
import sys
import glob
import importlib
import webbrowser
from accerciser.plugin import ViewportPlugin
from accerciser.i18n import _, N_, DOMAIN
import pyatspi

UI_FILE = os.path.join(os.path.dirname(__file__), 'validate.ui')
USER_SCHEMA_PATH = os.path.join(GLib.get_user_data_dir(), 'accerciser',
                                'plugindata', 'validate')
SYS_SCHEMA_PATH = os.path.join(sys.prefix, 'share', 'accerciser',
                               'plugindata', 'validate')
VALIDATORS = {}
SCHEMA_METADATA = {}

# method to make use of metaclasses on both python 2.x and 3.x
#
def with_metaclass(meta, *bases):
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__
        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})

class ValidatorManager(type):
  '''
  Metaclass that tracks all validator subclasses imported by the plug-in.
  Used to update a list of validator schemas at runtime.
  '''
  def __init__(cls, name, bases, dct):
    '''
    Build the class as usual, but keep track of its name and one instance.
    '''
    super(ValidatorManager, cls).__init__(name, bases, dct)
    if name != 'Validator':
      # don't count the base class
      names = cls.__module__.split('.')
      VALIDATORS.setdefault(names[-1], []).append(cls())

  @staticmethod
  def loadSchemas():
    '''
    Loads all schema files from well known locations.
    '''
    for path in [USER_SCHEMA_PATH, SYS_SCHEMA_PATH]:
      for fn in glob.glob(os.path.join(path, '*.py')):
        module_name = os.path.basename(fn)[:-3]
        spec = importlib.util.spec_from_file_location(module_name, fn)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        try:
          # try to get descriptive fields from the module
          SCHEMA_METADATA[module_name] = module.__metadata__
        except AttributeError:
          # default to usinf file name as description
          SCHEMA_METADATA[module_name] = {'name' : module,
                                          'description' : _('No description')}

  @staticmethod
  def getValidators(name):
    '''
    Gets all validator classes within a schema.

    @param name: Name of a schema
    @return: List of Validator objects
    @raise KeyError: When the schema is not known
    '''
    return VALIDATORS[name]

  @staticmethod
  def listSchemas():
    '''
    Gets a list of all available schema names.

    @return: List of string names
    '''
    return list(VALIDATORS.keys())

  @staticmethod
  def getSchemaMetadata(name):
    '''
    Gets information about a schema.

    @param name: Name of the schema
    @return: Dictionary of 'name', 'description', etc.
    '''
    return SCHEMA_METADATA[name]

class Validator(with_metaclass(ValidatorManager, object)):
  '''
  Base class for all validators. Defines the public interface used by the
  plug-in controller/view to generate validation reports.
  '''
  def __init__(self):
    pass

  def condition(self, acc):
    '''
    Checks if this validator is fit to test the given accessible. For instance,
    a test for table properties is not applicable to buttons.

    @param acc: Accessible to test
    @return: True to validate, False to avoid testing
    @raise Exception: Same as returning False
    '''
    return True

  def before(self, acc, state, view):
    '''
    Tests the accessible before testing its descendants.

    @param acc: Accessible to test
    @param state: Dictionary containing arbitrary state that will be passed
      among validators and be made available in both before and after passes
    @param view: View object to use to log test results
    '''
    pass

  def after(self, acc, state, view):
    '''
    Tests the accessible after testing its descendants.

    @param acc: Accessible to test
    @param state: Dictionary containing arbitrary state that will be passed
      among validators and be made available in both before and after passes
    @param view: View object to use to log test results
    '''
    pass

class ValidatorViewport(ViewportPlugin):
  '''
  Validator UI. Key feature is a table showing the results of a validation
  run on some accessible and its descendants.

  @ivar main_xml: gtk builder parsed XML definition
  @ivar report: Report table
  @ivar progress: Activity bar
  @ivar validate: Validation button
  @ivar save: Save report button
  @ivar schema: Schema combobox
  @ivar vals: All validators for the selected schema
  @ivar walk: Generator for the current validation walk through the hierarchy
  '''
  plugin_name = N_('AT-SPI Validator')
  plugin_name_localized = _(plugin_name)
  plugin_description = N_('Validates application accessibility')

  # keep track of when a file is being written
  write_in_progress = False

  def init(self):
    '''
    Loads the glade UI definition and initializes it.
    '''
    # load all schemas
    ValidatorManager.loadSchemas()

    # validators and walk generator
    self.vals = None
    self.walk = None
    # help url for last selected
    self.url = None

    self.main_xml = gtk.Builder()
    self.main_xml.set_translation_domain(DOMAIN)
    self.main_xml.add_from_file(UI_FILE)
    frame = self.main_xml.get_object('main vbox')
    self.plugin_area.add(frame)
    self.report = self.main_xml.get_object('report table')
    self.progress = self.main_xml.get_object('progress bar')
    self.validate = self.main_xml.get_object('validate button')
    self.help = self.main_xml.get_object('help button')
    self.save = self.main_xml.get_object('save button')
    self.clear = self.main_xml.get_object('clear button')
    self.schema = self.main_xml.get_object('schema combo')
    self.validator_buffer = gtk.TextBuffer()
    self.idle_validate_id = None

    # model for the combobox
    model = gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
    self.schema.set_model(model)

    # append all schema names/descriptions
    vm = ValidatorManager
    for name in vm.listSchemas():
      d = vm.getSchemaMetadata(name)
      model.append(['%s - %s' % (d['name'], d['description']), name])
    self.schema.set_active(0)

    # model for the report
    model = gtk.ListStore(str, str, object, str)
    self.report.set_model(model)

    # schema cell renderer
    cell = gtk.CellRendererText()
    self.schema.pack_start(cell, True)
    self.schema.add_attribute(cell, 'text', 0)

    # log level column
    col = gtk.TreeViewColumn(_('Level'))
    rend = gtk.CellRendererText()
    col.pack_start(rend, True)
    col.add_attribute(rend, 'text', 0)
    self.report.append_column(col)
    # description column
    rend = gtk.CellRendererText()
    col = gtk.TreeViewColumn(_('Description'), rend, text=1)
    self.report.append_column(col)

    # set progress bar to zero initially
    self.progress.set_fraction(0.0)

    self.main_xml.connect_signals(self)
    self.show_all()

  def onAccChanged(self, acc):
    '''
    Stops validation if the accessible hierarchy changes.

    @param acc: The accessible that changed
    '''
    if self.walk is not None:
      self._stopValidate()

  def _onValidate(self, widget):
    '''
    Starts or stops a validation run.

    @param widget: The validate button
    '''
    if widget.get_active():
      self._startValidate()
    else:
      self._stopValidate()

  def _writeFile(self):
    '''
    Save the report from the report model to disk in a temporary location.
    Close the file when finished.
    '''
    if self.write_in_progress:
      # if we have finished writing to the file
      if self.curr_file_row == self.n_report_rows:
        self.save_to.close()
        self._stopSave()
        return False
    else:
      # set up the file to be written
      self._startSave()
      self.save.set_sensitive(False)
      report_store = self.report.get_model()
      # create list of lists containing column values
      self.row_values = [[row[0], row[1], row[2], row[3]] for row in report_store]
      self.n_report_rows = len(self.row_values)
      return True

    remaining_rows = self.n_report_rows - self.curr_file_row
    n_rows_to_write = 5
    if n_rows_to_write > remaining_rows:
      n_rows_to_write = remaining_rows

    file_str_list = [] # list to store strings to be written to file
    start = self.curr_file_row
    end = (self.curr_file_row + n_rows_to_write)
    for i in range(start, end):
      val = self.row_values[i]
      # add level to buffer
      file_str_list.append("%s: %s\n" % (_('Level'), val[0]))
      # add description to buffer
      file_str_list.append("%s: %s\n" % (_('Description'), val[1]))
      # add accessible's name to buffer
      file_str_list.append("%s: %s\n" % (_('Name'), val[2].name))
      # add accessible's role to buffer
      file_str_list.append("%s: %s\n" % (_('Role'), val[2].getRoleName()))
      # add url role to buffer
      file_str_list.append("%s: %s\n\n" % (_('Hyperlink'), val[3]))
      self.curr_file_row += 1

    self.save_to.write(''.join(file_str_list))

    return True

  def _onSave(self, button):
    '''
    Save the report from the report model to disk

    @param button: Save button
    '''
    save_dialog = gtk.FileChooserNative.new(
      _('Save validator output'),
      self.get_toplevel(),
      gtk.FileChooserAction.SAVE,
      _('_OK'),
      _('_Cancel'))
    save_dialog.set_do_overwrite_confirmation(True)
    response = save_dialog.run()
    if response == gtk.ResponseType.ACCEPT:
      self.save_to = open(save_dialog.get_filename(), 'w')
      GLib.idle_add(self._writeFile)
    save_dialog.destroy()

  def _onClear(self, button):
    '''
    Clear the report from the report model

    @param button: Clear button
    '''
    self.report.get_model().clear()
    self.validator_buffer.set_text('')
    self.save.set_sensitive(False)
    self.clear.set_sensitive(False)

  def _onHelp(self, button):
    '''
    Open a help URL in the system web browser.

    @param button: Help button
    '''
    webbrowser.open(self.url)

  def _onSaveIdle(self):
    '''
    Move the progress bar
    '''
    self.progress.pulse()
    return True

  def _startSave(self):
    '''
    Starts a save by settting up an idle callback after initializing progress
    bar.
    '''
    # set variables for writing report to file
    self.write_in_progress = True
    self._setDefaultSaveVars()
    # register an idle callback
    self.idle_save_id = GLib.idle_add(self._onSaveIdle)
    self.progress.set_text(_('Saving'))
    # disable controls
    self.validate.set_sensitive(False)
    self.save.set_sensitive(False)

  def _stopSave(self):
    '''
    Stops a save by disabling the idle callback and restoring the various UI
    components to their enabled states.
    '''
    # stop callbacks
    GLib.source_remove(self.idle_save_id)
    # reset progress
    self.progress.set_fraction(0.0)
    self.progress.set_text(_('Idle'))
    # enable other controls
    self.validate.set_sensitive(True)
    self.save.set_sensitive(True)
    self.save.set_sensitive(True)
    # reset variables for writing report to file
    self._setDefaultSaveVars()
    self.write_in_progress = False

  def _setDefaultSaveVars(self):
    '''
    Ready appropriate variables for a save
    '''
    self.curr_file_row = 0
    self.n_report_rows = 0
    self.row_values = []

  def _startValidate(self):
    '''
    Starts a validation by settting up an idle callback after initializing the
    report table and progress bar. Gets all validators for the selected schema.
    '''
    # clear the report
    self.report.get_model().clear()
    # get the validators
    index = self.schema.get_active()
    if index == -1:
      self.validate.set_active(False)
      return
    row = self.schema.get_model()[index]
    self.vals = ValidatorManager.getValidators(row[1])
    # build a new state dict
    state = {}
    # build our walk generator
    self.walk = self._traverse(self.acc, state)
    # register an idle callback
    self.idle_validate_id = GLib.idle_add(self._onValidateIdle)
    self.progress.set_text(_('Validating'))
    # disable controls
    self.schema.set_sensitive(False)
    self.help.set_sensitive(False)
    self.save.set_sensitive(False)
    self.clear.set_sensitive(False)

  def _stopValidate(self):
    '''
    Stops a validation run by disabling the idle callback and restoring the
    various UI components to their enabled states.
    '''
    if self.idle_validate_id == None:
      return
    # stop callbacks
    GLib.source_remove(self.idle_validate_id)
    # destroy generator
    self.walk = None
    # reset progress
    self.progress.set_fraction(0.0)
    self.progress.set_text(_('Idle'))
    # depress validate
    self.validate.set_active(False)
    # enable other controls
    self.schema.set_sensitive(True)
    self.help.set_sensitive(True)
    self.save.set_sensitive(True)
    self.clear.set_sensitive(True)

  def _onValidateIdle(self):
    '''
    Tests one accessible at a time on each idle callback by advancing the
    walk generator.
    '''
    try:
      # generate the next accessible to validate
      next(self.walk)
    except StopIteration:
      # nothing left to validate, so stop
      self._stopValidate()
      return False
    # pule the progress bar
    self.progress.pulse()
    # yes, we want more callbacks
    return True

  def _traverse(self, acc, state):
    '''
    Generates accessibles in a two-pass traversal of the subtree rooted at
    the accessible selected at the time the validation starts. Accessibles are
    tested first before their descendants (before pass) and then after all of
    their descendants (after pass).

    @param acc: Accessible root of some subtree in the walk
    @param state: State object used by validators to share information
    '''
    # start the walk generator
    gen_child = self._genAccessible(acc, state)
    while 1:
      try:
        # get one child
        child = next(gen_child)
      except StopIteration as e:
        break
      # recurse
      gen_traverse = self._traverse(child, state)
      while 1:
        # yield before continuing processing
        yield None
        try:
          # get one descendant
          next(gen_traverse)
        except StopIteration:
          break

  def _genAccessible(self, acc, state):
    '''
    Tests the given accessible in the before pass if its test condition is
    satisfied. Then generates all of its children. Finally, tests the original
    accessible in the after pass if its test condition is satisfied.

    @param acc: Accessible to test
    @param state: State object used by validators to share information
    '''
    # run before() methods on all validators
    self._runValidators(acc, state, True)
    # generate all children, but only if acc doesn't manage descendants
    if not acc.getState().contains(pyatspi.constants.STATE_MANAGES_DESCENDANTS):
      for i in range(acc.childCount):
        child = acc.getChildAtIndex(i)
        yield child
    # run after methods on all validators
    self._runValidators(acc, state, False)

  def _runValidators(self, acc, state, before):
    '''
    Runs all validators on a single accessible. If the validator condition is
    true, either executes the before() or after() method on the validator
    depending on the param 'before' passed to this method.

    @param acc: Accessible to test
    @param state: State object used by validators to share information
    @param before: True to execute before method, false to execute after
    '''
    for val in self.vals:
      try:
        ok = val.condition(acc)
      except Exception:
        pass
      else:
        if ok:
          try:
            if before:
              val.before(acc, state, self)
            else:
              val.after(acc, state, self)
          except Exception as e:
            self._exceptionError(acc, e)

  def _onCursorChanged(self, report):
    '''
    Enables or disables the help button based on whether an item has help or
    not.

    @param report: Report table
    '''
    selection = report.get_selection()
    model, iter = selection.get_selected()
    if iter:
      url = model[iter][3]
      self.help.set_sensitive(len(url))
      self.url = url

  def _onActivateRow(self, report, iter, col):
    '''
    Updates the Accerciser tree to show an accessible when a report entry is
    selected.

    @param report: Report table
    @param iter: Tree table iterator
    @param col: Tree table column
    '''
    selection = report.get_selection()
    model, iter = selection.get_selected()
    if iter:
      acc = model[iter][2]
      if acc:
        self.node.update(acc)

  def _exceptionError(self, acc, ex):
    '''
    Logs an unexpected exception that occurred during execution of a validator.

    @param acc: Accessible under test when the exception occurred
    @param ex: The exception
    '''
    info = traceback.extract_tb(sys.exc_info()[2])
    text = '%s (%d): %s' % (os.path.basename(info[-1][0]), info[-1][1], ex)
    self.report.get_model().append([_('EXCEPT'), text, acc, ''])

  def error(self, text, acc, url=''):
    '''
    Used by validators to log messages for accessibility problems that have to
    be fixed.
    '''
    level = _('ERROR')
    self.report.get_model().append([level, text, acc, url])

  def warn(self, text, acc, url=''):
    '''
    Used by validators to log warning messages for accessibility problems that
    should be fixed, but are not critical.
    '''
    level = _('WARN')
    self.report.get_model().append([level, text, acc, url])

  def info(self, text, acc, url=''):
    '''
    Used by validators to log informational messages.
    '''
    level = _('INFO')
    self.report.get_model().append([level, text, acc, url])

  def debug(self, text, acc, url=''):
    '''
    Used by validators to log debug messages.
    '''
    level = _('DEBUG')
    self.report.get_model().append([level, text, acc, url])

  def close(self):
    '''
    Things to do before the plugin closes.
    '''
    # don't close the plugin until we have finished writing
    while True:
      if not self.write_in_progress:
        break
      gtk.main_iteration_do(True)


